#!/usr/bin/env python 
"""The module contain functions that can be used for developing & testing other functions of the library 
"""
# import the modules 
import pandas as pd
from Bio import SeqIO
import numpy as np
from typing import List , Set, Dict 
import random
from IPTK.Classes.Experiment import Experiment
from IPTK.Classes.Proband import Proband
from IPTK.Classes.HLASet import HLASet 
from IPTK.Classes.Tissue import Tissue
from IPTK.Classes.Database import SeqDB
from IPTK.Utils.UtilityFunction import generate_random_name
# define the functions
def simulate_an_experimental_ident_table_from_fasta(path2load: str, num_pep: int, num_prot: int )->pd.DataFrame:
    """Simulate an IP identification table from a fasta file. Please Note,  if the reminder of num_pep over num_prot does not equal to zero,
    the floor of this ratio will be used to sample peptides from each proteins 

    :param path2load:  The path to load the Fasta files 
    :type path2load: str
    :param num_pep: The number of peptides in the tables 
    :type num_pep: int
    :param num_prot: The number of UNIQUE proteins in the table
    :type num_prot: int
    :raises ValueError: if number of proteins or number of peptide is zero 
    :return: an identification table 
    :rtype: pd.DataFrame
    """
    # load the fasta files
    # check that number of proteins & number of peptides does not equal zero 
    if not num_pep or not num_prot:
        raise ValueError('number of peptides or number of proteins can not be zero')
    fasta_gene=SeqIO.parse(path2load,'fasta')
    proteins_dict=dict()
    # load the desired number of proteins
    idx=0 
    while idx < num_prot:
        temp_seq=next(fasta_gene)
        if len(str(temp_seq.seq)) <50:
            idx-=1
            continue
        proteins_dict[temp_seq.id.split('|')[1]]=str(temp_seq.seq)
        idx+=1
    # get the average number of peptides from proteins 
    num_peptides_=int(np.floor(num_pep/num_prot))
    # generate the results 
    peptides,proteins,start_idx,end_idx=[],[],[],[]
    proteins_names=list(proteins_dict.keys())
    for prot_idx in range(num_prot):
        protine_seq=proteins_dict[proteins_names[prot_idx]]
        for idx in range(num_peptides_):
            pep_length=int(np.random.normal(15,3))
            anchor_point=np.random.randint(low=pep_length,high=len(protine_seq)-pep_length)
            peptides.append(protine_seq[anchor_point:anchor_point+pep_length])
            proteins.append(proteins_names[prot_idx])
            start_idx.append(anchor_point)
            end_idx.append(anchor_point+pep_length)
    # return the results as a data.frame 
    res=pd.DataFrame({
        'peptides':peptides,
        'proteins':proteins,
        'start_index':start_idx,
        'end_index':end_idx
    })
    return res

def simulate_an_expression_table(num_transcripts: int = 100) ->pd.DataFrame:
    """ create a dummy expression table to be used for testing and developing Tissue based classes 

    :param num_transcripts: The number of transcripts that shall be contained in the transcript, defaults to 100
    :type num_transcripts: int, optional
    :raises ValueError: incase number of transcripts is 0
    :return: a stimulated expression label 
    :rtype: pd.DataFrame
    """
    # check that the input is a positive integer 
    if num_transcripts<=0:
        raise ValueError(f'Number of Transcripts must be bigger than zero')
    # simulate the data 
    trans_id, prot_name, trans_value= [],[],[]
    for _ in range(num_transcripts):
        trans_id.append('ENS_'+generate_random_name(12))
        prot_name.append(generate_random_name(12))
        trans_value.append(np.random.normal(25,10))
    # create the data table 
    return pd.DataFrame(
        {
            'Transcript_id':trans_id,
            'Protein_name': prot_name,
            'Expression_value':trans_value
        }
    )

def simulate_random_experiment(alleles: List[str],  path2fasta: str,  tissue_name: str='TEST_TISSUE',
        num_pep: int = 10, num_prot: int = 5, proband_name: str = None )->Experiment:
    """ Simulate a random experiment objects 

    :param alleles: a list of alleles names. 
    :type alleles: List[str]
    :param path2fasta: The path to load the database objects 
    :type path2fasta: str
    :param tissue_name: the name of the tissue, defaults to 'TEST_TISSUE'
    :type tissue_name: str, optional
    :param num_pep: the number of peptides in the table, defaults to 10
    :type num_pep: int, optional
    :param num_prot: number of proteins, defaults to 5
    :type num_prot: int, optional
    :param proband_name: The name of the Proband, defaults to None
    :type proband_name: str, optional
    :return: A simulated experimental object 
    :rtype: Experiment
    """
    if proband_name is None:
        proband_name=generate_random_name(12)
    
    proband: Proband = Proband(name=proband_name)
    hla_set: HLASet = HLASet(alleles)
    ident_table: pd.DataFrame =simulate_an_experimental_ident_table_from_fasta(path2fasta,num_pep,num_pep)
    # to be upgraded to the new version of the Tissue class 
    tissue: Tissue = Tissue(tissue_name,simulate_an_expression_table(num_transcripts=1000),
            simulate_an_expression_table(num_transcripts=100))
    database: SeqDB = SeqDB(path2fasta)
    return Experiment(proband=proband,hla_set=hla_set,tissue=tissue,database=database,
    ident_table=ident_table)

def generate_random_peptide_seq(peptide_length:int, num_peptides: int)->List[str]:
    """generate a list of random peptides for testing and development purposes.
    
    :param peptide_length: The peptide length 
    :type peptide_length: int
    :param num_peptides: The number of peptides in the generate list  
    :type num_peptides: int
    :return: a list of random peptides 
    :rtype: List[str]
    """
    # define amino acids 
    amino_acids=['G','A','L','M','F','W','K','Q','E','S','P','V','I','C','Y','H','R','N','D','T']
    results=[]
    # generate the sequences
    for _ in range(num_peptides):
        results.append(
            ''.join([random.choice(amino_acids) for _ in range(peptide_length)])
        )
    # return the results 
    return results

def simulate_mapped_array_list(min_len: int = 20, max_len: int = 100, num_elem: int = 100)->List[np.ndarray]:
    """Simulate a list of mapped arrays proteins to be used for development purposes 
    
    :param min_len: the minmum length of the protein, defaults to 20
    :type min_len: int, optional
    :param max_len: the maximum length for the protein, defaults to 100
    :type max_len: int, optional
    :param num_elem: the number of arrays in the protein, defaults to 100
    :type num_elem: int, optional
    :return: A list of simulated NumPy array that represent protein peptide coverage
    :rtype: List[np.ndarray]
    """
    # allocate a list to hold the results 
    results: List[np.ndarray] = []
    # fill the list with arrays: 
    for _ in range(num_elem):
        array_length: int = np.random.randint(low=min_len,high=max_len)
        results.append(np.random.randint(low=0,high=2,size=(1,array_length)))
    # return the results 
    return results


#!/usr/bin/env python 
"""A submodule that contain function to map different database keys 
"""
# load the modules: 
import urllib
from typing import List, Dict
import pandas as pd 
# define the mapping function 
def map_from_uniprot_pdb(uniprots: List[str])-> pd.DataFrame:
    """map from uniprot id to protein data bank identifiers

    :param uniprots: a list of uniprot IDs 
    :type uniprots: List[str]
    :return: A table that contain the mapping between each uniprot and its corresponding PDB ID/IDs  
    :rtype: pd.DataFrame
    """
    url: str ='https://www.uniprot.org/uploadlists/'
    # define the query parameters 
    q_params: Dict[str, str]={
        'from': 'ACC+ID', 
        'to': 'PDB_ID',
        'format': 'tab',
        'query': ' '.join(uniprots)
    }
    data: bytes =urllib.parse.urlencode(q_params).encode('utf-8')
    request: urllib.request.Request = urllib.request.Request(url,data)
    # read the request
    with urllib.request.urlopen(request) as input_file: 
        results: str =input_file.read().decode('utf-8')
    # parse the resulting strings 
    mapped_pairs: List[str] = results.split('\n')
    # pop the first element as it contain the words from and to 
    mapped_pairs.pop(0)
    # allocate to lists to hold the results 
    unitpot_ids: List[str] = []
    pdb_ids: List[str] = []
    # parse the results 
    for pair in mapped_pairs:
        temp_lists: List[str] = pair.split('\t')
        if len(temp_lists) ==2:
            unitpot_ids.append(temp_lists[0])
            pdb_ids.append(temp_lists[1])
    # combine the data into a dataframe 
    results: pd.DataFrame = pd.DataFrame({
        'Uniprot-ID':unitpot_ids,
        'PDB':pdb_ids
    })
    # return the results 
    return results

def map_from_uniprot_gene(uniprots: List[str])->pd.DataFrame: 
    """ map from uniprot id to ensemble gene ids

    :param uniprots: a list of uniprot IDs 
    :type uniprots: List[str]
    :return: A table that contain the mapping between each uniprot and its corresponding Gene ID/IDs  
    :rtype: pd.DataFrame
    """
    url: str ='https://www.uniprot.org/uploadlists/'
    # define the query parameters 
    q_params: Dict[str, str]={
        'from': 'ACC+ID', 
        'to': 'ENSEMBL_ID',
        'format': 'tab',
        'query': ' '.join(uniprots)
    }
    data: bytes =urllib.parse.urlencode(q_params).encode('utf-8')
    request: urllib.request.Request = urllib.request.Request(url,data)
    # read the request
    with urllib.request.urlopen(request) as input_file: 
        results: str =input_file.read().decode('utf-8')
    # parse the resulting strings 
    mapped_pairs: List[str] = results.split('\n')
    # pop the first element as it contain the words from and to 
    mapped_pairs.pop(0)
    # allocate to lists to hold the results 
    unitpot_ids: List[str] = []
    ensemble_ids: List[str] = []
    # parse the results 
    for pair in mapped_pairs:
        temp_lists: List[str] = pair.split('\t')
        if len(temp_lists) ==2:
            unitpot_ids.append(temp_lists[0])
            ensemble_ids.append(temp_lists[1])
    # combine the data into a dataframe 
    results: pd.DataFrame = pd.DataFrame({
        'Uniprot-ID':unitpot_ids,
        'Gene-ID':ensemble_ids
    })
    # return the results 
    return results


#!/usr/bin/env Python 
"""Parse different user inputs into a standard format/tables used by the library. 
"""
# load the models 
import pandas as pd 
import numpy as np 
from typing import Dict, List
import os  
from Bio import SeqIO
from Bio.PDB import PDBList
from pyteomics.mztab import MzTab
from pyteomics import pepxml, auxiliary
from pyteomics.openms import idxml
# define the functions of the modules

def load_identification_table(input_path: str, sep:str) -> pd.DataFrame:
    """load & process an identification table 
    
    :param input_path: the path to the identification table. With the following columns: peptides which hold the identified
	peptide sequence, protein which hold the identified protein sequence, start_index, and end_index where
	the last two columns define the position of the peptide in the parent protein. 
    :type input_path: str
    :param sep: The separator to parse the provided table. 
    :type sep: str
    :rtype: pd.DataFrame
    """
	# load the table 
    try:
	    table: pd.DataFrame = pd.read_csv(input_path, sep=sep)
    except Exception as exp:
        raise IOError(f'Loading the input table from the provide path; {input_path} gave rise to the following IOException: {exp}')
    # assert that it has the correct shape 
    if table.shape[1]!= 4:
        raise ValueError(f'The provided table does not have the expected number of columns, Expected 4 columns, however, {table.shape} columns were found')
    # set the column names
    table.columns=['peptide','protein','start_index','end_index']
    # return the table 
    return table 

def parse_mzTab_to_identification_table(path2mzTab: str, path2fastaDB: str,
    fasta_reader_param: Dict[str,str]={'filter_decoy':True, 'decoy_string':'DECOY' },
    remove_if_not_matched: bool = True)->pd.DataFrame:
    """parse a user provided mzTab to an identification table 

    :param path2mzTab: the path to the input mzTab file
    :type path2mzTab: str
    :param path2fastaDB: the path to a fasta file to obtain the protein sequences 
    :type path2fastaDB: str
    :param fasta_reader_param: A dict of parameters for controlling the behavior of the fasta reader , defaults to {'filter_decoy':True, 'decoy_string':'DECOY' }
    :type fasta_reader_param: Dict[str,str], optional
    :param remove_if_not_matched: remove the peptide if it could not be matched to the parent protein, defaults to True
    :type remove_if_not_matched: bool, optional
    :raises IOError: if the mztab file could not be open and loaded or if the fasta database could not be read
    :raises KeyError: if a protein id defined in the mzTab file could not be extracted from a matched sequence database
    :raises ValueError: if the peptide can not be mapped to the identified protein 
    :return: the identification table 
    :rtype: pd.DataFrame
    """
    # load the files 
    try: 
        input_file: MzTab =MzTab(path2mzTab)
    except Exception as exp:
        raise IOError(f'While parsing your input mzTab file: {path2mzTab}, the following error was encountered: {exp}')
    # load the fasta files and extract the accession of the proteins 
    try:
        sequence_dict: Dict[str,str]= fasta2dict(path2fastaDB, **fasta_reader_param)
    except Exception as exp:
        raise IOError(f'While parsing your input fasta file: {path2fastaDB}, the following error was encountered: {exp}')
    # get the peptide tables
    peptides_table: pd.DataFrame = input_file.peptide_table
    # construct the identification table 
    peptides: List[str] = peptides_table.sequence.tolist()
    protein_acc: List[str]= [acc.split('|')[1] for acc in peptides_table.accession.tolist()]
    start_index: List[int] = []
    end_index: List[int] = []
    # fill extract the start and end-index information from the library 
    for idx in range(len(protein_acc)):
        # get the protein sequence 
        try:
            prot_seq: str = sequence_dict[protein_acc[idx]]
        except KeyError as exp: 
            raise KeyError(f'Database mismatch, the current protein accession: {protein_acc[idx]} is not defined in the provided sequence database')
       # get the index of the protein sequence
        try: 
            if '(' in peptides[idx]: # that is there sequence modifications in the sequence 
                temp_peptide=peptides[idx] # that is there sequence modifications in the sequence 
                while '(' in temp_peptide or ')' in temp_peptide: 
                    pre_seq=temp_peptide.split('(')[0]
                    post_seq=")".join(temp_peptide.split(')')[1:])
                    temp_peptide=pre_seq+post_seq
                start_index.append(prot_seq.index(temp_peptide))
                peptide_len=len(temp_peptide)
            else: 
                start_index.append(prot_seq.index(peptides[idx]))
                peptide_len=len(peptides[idx])
        except ValueError as exp:
            if remove_if_not_matched:
                start_index.append(-1) #  add a placeholder value that will be dropped later 
            else: 
                raise ValueError(f'Peptide sequence: {peptides[idx]} could not be extracted from protein sequence: {prot_seq} with accession: {protein_acc[idx]}')
        # add the end index 
        end_index.append(start_index[idx]+peptide_len)
    # build the data frame 
    ident_table: pd.DataFrame = pd.DataFrame({
        'peptide': peptides,
        'protein': protein_acc,
        'start_index':start_index,
        'end_index':end_index
    })
    # return the results 
    return ident_table
     
def parse_xml_based_format_to_identification_table(path2XML_file: str, path2fastaDB: str,
    decoy_prefix: str ='DECOY', is_idXML: bool = False, 
    fasta_reader_param: Dict[str,str]={'filter_decoy':True, 'decoy_string':'DECOY' }, 
    remove_if_not_matched: bool = True)->pd.DataFrame:
    """parse either a pepXML or an idXML file to generate an identification table , 

    :param path2XML_file: The path to the input pepXML files
    :type path2XML_file: str
    :param path2fastaDB: The path to a fasta sequence database to obtain the protein sequences
    :type path2fastaDB: str
    :param decoy_prefix: the prefix of the decoy sequences, default is DECOY
    :type decoy_prefix: str, optional
    :param is_idXML: Whether or not the provided file is an idXML, default is false which assume the provided file is a pepXML file, defaults to False
    :type is_idXML: bool, optional
    :param fasta_reader_param: A dict of parameters for controlling the behavior of the fasta reader, defaults to {'filter_decoy':True, 'decoy_string':'DECOY' }
    :type fasta_reader_param: Dict[str,str], optional
    :param remove_if_not_matched: remove the peptide if it could not be matched to the parent protein, defaults to True
    :type remove_if_not_matched: bool, optional
    :raises IOError: if the fasta database could not be open 
    :raises ValueError: if the XML file can not be open 
    :raises KeyError: if a protein id defined in the mzTab file could not be extracted from a matched sequence database
    :raises ValueError: if the peptide can not be mapped to the identified protein 
    :return: the identification table 
    :rtype: pd.DataFrame
    """
    # load the fasta files and extract the accession of the proteins 
    try:
        sequence_dict: Dict[str,str]= fasta2dict(path2fastaDB,**fasta_reader_param)
    except Exception as exp:
        raise IOError(f'While parsing your input fasta file: {path2fastaDB}, the following error was encountered: {exp}')
    # parse that the file exists:  
    if not os.path.exists(path2XML_file):
        raise ValueError(f'The provided path: {path2XML_file} does not exist!')
    # allocate a list to hold peptide and protein list 
    peptides: List[str] = []
    protein_acc: List[str] = []
    #  parse the XML file 
    if is_idXML: 
        with idxml.IDXML(path2XML_file) as reader:
            for elem in reader:
                for hit in elem['PeptideHit']:
                    for prot in hit['protein']: 
                        if decoy_prefix not in prot['accession']:
                            peptides.append(hit['sequence'])
                            if '|' in prot['accession']: 
                                protein_acc.append(prot['accession'].split('|')[1])
                            else: 
                                protein_acc.append(prot['accession'])
    else: 
        with pepxml.read(path2XML_file) as reader: 
            for elem in reader:
                for hit in elem['search_hit']:
                    for protein in hit['proteins']: 
                        if decoy_prefix not in protein['protein']:
                            peptides.append(hit['peptide'])
                            protein_acc.append(protein['protein'].split('|')[1])
    # extract the start and end index of the peptides from the parent proteins 
    start_index: List[int] = []
    end_index: List[int] = []
    # fill extract the start and end-index information from the library 
    for idx in range(len(protein_acc)):
        # get the protein sequence 
        try:
            prot_seq: str = sequence_dict[protein_acc[idx]]
        except KeyError as exp: 
            raise KeyError(f'Database mismatch, the current protein accession: {protein_acc[idx]} is not defined in the provided sequence database')
        # get the index of the protein sequence
        try: 
            if '(' in peptides[idx]: # that is there sequence modifications in the sequence 
                temp_peptide=peptides[idx] # that is there sequence modifications in the sequence 
                while '(' in temp_peptide or ')' in temp_peptide: 
                    pre_seq=temp_peptide.split('(')[0]
                    post_seq=")".join(temp_peptide.split(')')[1:])
                    temp_peptide=pre_seq+post_seq
                start_index.append(prot_seq.index(temp_peptide))
                peptide_len=len(temp_peptide)
            else: 
                start_index.append(prot_seq.index(peptides[idx]))
                peptide_len=len(peptides[idx])
        except ValueError as exp:
            if remove_if_not_matched:
                start_index.append(-1) #  add a placeholder value that will be dropped later 
            else: 
                raise ValueError(f'Peptide sequence: {peptides[idx]} could not be extracted from protein sequence: {prot_seq} with accession: {protein_acc[idx]}')
        # add the end index 
        end_index.append(start_index[idx]+peptide_len)
    # build the data frame 
    ident_table: pd.DataFrame = pd.DataFrame({
        'peptide': peptides,
        'protein': protein_acc,
        'start_index':start_index,
        'end_index':end_index
    })
    ident_table=ident_table.loc[ident_table.iloc[:,2]!=-1,:] # filter the non-matched peptides 
    # return the results 
    return ident_table


def parse_text_table(path2file: str, 
                    path2fastaDB: str,
                    sep=',',
                    fasta_reader_param: Dict[str,str]={'filter_decoy':True, 'decoy_string':'DECOY' }, 
                    seq_column: str = 'Sequence',
                    accession_column: str = 'Protein Accessions',
                    protein_group_sep: str = ';',
                    remove_protein_version: bool = True,
                    remove_if_not_matched: bool = True
            )->pd.DataFrame:
    """Parse a user defined table to extract the columns containing the identification table 
    
    :param path2file: The path to load the CSV file holding the results 
    :type path2file: str
    :param path2fastaDB: The path to a fasta sequence database to obtain the protein sequences
    :type path2fastaDB: str
    :param sep: The table separators, defaults to ','
    :type sep: str, optional
    :param fasta_reader_param: a dict of parameters for controlling the behavior of the fasta reader, defaults to {'filter_decoy':True, 'decoy_string':'DECOY' }
    :type fasta_reader_param: Dict[str,str], optional
    :param seq_column: The name of the columns containing the peptide sequence, defaults to 'Sequence'
    :type seq_column: str, optional
    :param accession_column: The name of the column containing the protein accession, defaults to 'Protein Accessions'
    :type accession_column: str, optional
    :param protein_group_sep: The separator for the protein group, defaults to ';'
    :type protein_group_sep: str, optional
    :param remove_protein_version: A bool if true strip the version number from the protein , defaults to True
    :type remove_protein_version: bool, optional
    :param remove_if_not_matched: remove the peptide if it could not be matched to the parent protein, defaults to True
    :type remove_if_not_matched: bool, optional
    :raises IOError: Incase either the sequences database or the identification table can not be open and loaded
    :raises KeyError: In case the provided column names not in the provided identification table. 
    :raises KeyError: Incase the protein sequence can not be extract from the sequence database 
    :raises ValueError: incase the peptide could not be located in the protein sequence 
    :return: an identification table
    :rtype: pd.DataFrame
    """
    # load the fasta files and extract the accession of the proteins 
    try:
        sequence_dict: Dict[str,str]= fasta2dict(path2fastaDB,**fasta_reader_param)
    except Exception as exp:
        raise IOError(f'While parsing your input fasta file: {path2fastaDB}, the following error was encountered: {exp}')
    # Load the table 
    try: 
        input_table: pd.DataFrame = pd.read_csv(path2file,sep=sep)
    except Exception as exp:
        raise IOError(f'while loading the input table: {path2file}, The following error was encountered: {exp}')
    # extract the columns of the table 
    if seq_column not in input_table.columns or accession_column not in input_table.columns :
        raise KeyError(f'The provided names for the peptides sequence: {seq_column} and/or the indexing column: {accession_column} and/or accession column: {accession_column} could not be found on the table')
    # allocate the lists
    peptides: List[str] = []
    protein_acc: List[str] = []
    start_index: List[int] = []
    end_index: List[int] = []
    # loop of the table to fill the lists 
    for _, row in input_table.iterrows():
        # check if the peptide is present in more than one protein 
        if protein_group_sep in row[accession_column]:
            accessions: List[str] =row[accession_column].split(protein_group_sep)
            for accession in accessions:
                # extract the macting protein sequence 
                accession=accession.strip(' ')
                try: 
                    protein_seq: str = sequence_dict[accession]
                except KeyError as exp:
                    if remove_if_not_matched:
                        accession=accession.split('-')[0]
                        try:
                            protein_seq= sequence_dict[accession]
                        except KeyError as exp: 
                            raise KeyError(f"Database mismatch, the current protein accession: {accession.strip(' ')} is not defined in the provided sequence database")
                    else: 
                        raise KeyError(f"Database mismatch, the current protein accession: {accession.strip(' ')} is not defined in the provided sequence database")
                peptides.append(row[seq_column])
                protein_acc.append(accession)
                # get the index 
                try: 
                    temp_start_idx: int =protein_seq.index(row[seq_column])
                except ValueError as exp:
                    if remove_if_not_matched: 
                        temp_start_idx=-1
                        pass
                    else: 
                        raise ValueError(f'Peptide sequence: {row[seq_column]} could not be extracted from protein sequence: {protein_seq}')
                start_index.append(temp_start_idx)
                end_index.append(temp_start_idx+len(row[seq_column]))
        else: 
            accession: int = row[accession_column].strip(' ')
            try: 
                protein_seq: str = sequence_dict[accession]
            except KeyError as exp:
                if remove_if_not_matched:
                    accession=accession.split('-')[0]
                    try:
                        protein_seq= sequence_dict[accession]
                    except KeyError as exp: 
                        raise KeyError(f"Database mismatch, the current protein accession: {accession} is not defined in the provided sequence database")
                else: 
                    raise KeyError(f"Database mismatch, the current protein accession: {accession} is not defined in the provided sequence database")
            peptides.append(row[seq_column])
            protein_acc.append(accession)
            # get the index 
            try: 
                temp_start_idx: int =protein_seq.index(row[seq_column])
            except ValueError as exp:
                if remove_if_not_matched: 
                    temp_start_idx=-1
                    pass
                else: 
                    raise ValueError(f'Peptide sequence: {row[seq_column]} could not be extracted from protein sequence: {protein_seq}')
            start_index.append(temp_start_idx)
            end_index.append(temp_start_idx+len(row[seq_column]))            
    # build the data frame 
    ident_table: pd.DataFrame = pd.DataFrame({
        'peptide': peptides,
        'protein': protein_acc,
        'start_index':start_index,
        'end_index':end_index
    })
    ident_table=ident_table.loc[ident_table.iloc[:,2]!=-1,:] # filter the non-matched peptides 
    # return the results 
    return ident_table 

def fasta2dict(path2fasta:str, filter_decoy: bool = True,
            decoy_string: str = 'DECOY')->Dict[str,str]: 
    """loads a fasta file and construct a dict object where ids are keys and sequences are the value

    :param path2fasta: The path to load the fasta file 
    :type path2fasta: str
    :param filter_decoy: A boolean of whether or not to filter the decoy sequences from the database, defaults to True
    :type filter_decoy: bool, optional
    :param decoy_string: The decoy database prefix, only valid incase filter_decoy is set to true, defaults to 'DECOY'
    :type decoy_string: str, optional
    :raises IOError: [description]
    :return: A dict where the protein ids are the keys and the protein sequences are the values 
    :rtype: Dict[str,str]
    """
    # check the path exists 
    if not os.path.exists(path2fasta):
        raise IOError(f"The provided path to fasta: {path2fasta} does not exits!")
    # load the fasta file 
    seq_gen: SeqIO.FastaIO.FastaIterator = SeqIO.parse(path2fasta,'fasta')
    results=dict()
    for seq in seq_gen: 
        if filter_decoy:
            if decoy_string in seq.id:
               continue
            else:
                if '|' in seq.id: 
                    results[seq.id.split('|')[1]]=str(seq.seq)
                else: 
                    results[seq.id]=str(seq.seq)
        else:
            if '|' in seq.id: 
                results[seq.id.split('|')[1]]=str(seq.seq)
            else: 
                results[seq.id]=str(seq.seq)
    # return the results 
    return results

def download_pdb_entry(prot_id: str) ->str: 
    """ Download the structure of a protein from protein databank as an mmCIF file. 

    :param prot_id: the protein id 
    :type prot_id: str
    :raises IOError: incase downloading and accessing the data failed 
    :return: the path to the downloaded file 
    :rtype: str
    """
    pdb_res=PDBList()
    try: 
        return pdb_res.retrieve_pdb_file(prot_id)
    except Exception as exp: 
        raise IOError(f'While Downloading the structure of protein: {prot_id}, the following error was encountered: {exp}')




#!/usr/bin/env python 
"""Contain a definition of commonly used types through the library 
"""
# load the modules 
from typing import List, Set, Tuple, Dict
import numpy as np
from IPTK.Classes.HLAChain import HLAChain
# define the types
Sequences = List[str]
MappedProtein = List[np.ndarray]
Proteins=Set[str]
Range=Tuple[int, int]
MappedProteins=Dict[str,MappedProtein]
FastaSet=Dict[str,str]
ProteinHits=List[str]
PlottingKeywards=Dict[str,str]
MappedProteinRepresentation=Dict[str,np.ndarray]
Index=List[int]
HLA_Chains=Dict[str,HLAChain]
Genes=List[str]
AlleleGroup=List[str]
ProteinGroup=List[str]
HLA_Names=List[str]
ProteinSource=Dict[str,str]
Organisms=List[str]



#!/usr/bin/env Python 
"""The module contains a collection of analysis functions that can be used by the methods of
the classes defined in the IPTK.Classes module. 
"""
# load the modules: 
import numpy as np 
import pandas as pd 
import subprocess as sp 
import os
from Bio.PDB import PDBList
from Bio.motifs.meme import Motif 
from typing import List, Callable, Dict, Set
from IPTK.IO import MEMEInterface as memeIF
from IPTK.IO import OutFunctions as out_func
from IPTK.Classes.Experiment import Experiment
from IPTK.Utils.UtilityFunction import check_peptide_made_of_std_20_aa
from IPTK.Utils.Mapping import map_from_uniprot_gene
from scipy.stats import pearsonr
from IPTK.Classes.Features import Features
# define some types 
Peptides=List[str]
Proteins=List[str]
## Define the peptide overlap 
def get_binnary_peptide_overlap(exp1:Experiment, exp2:Experiment)->Peptides:
    """compare the peptide overlap between two experimental objects.

    :param exp1: an instance of class Experiment.
    :type exp1: Experiment
    :param exp2: an instance of class Experiment.
    :type exp2: Experiment
    :return: a list of peptides that have been identified in both experiments.
    :rtype: Peptides
    """
    peptide_one=exp1.get_peptides()
    peptide_two=exp2.get_peptides()
    return list(peptide_one.intersection(peptide_two))
     
def get_binnary_protein_overlap(exp1:Experiment, exp2:Experiment)->Proteins:
    """compare the protein overlap between two experimental objects.

    :param exp1: an instance of class Experiment.
    :type exp1: Experiment
    :param exp2: an instance of class Experiment.
    :type exp2: Experiment
    :return: a list of proteins that have been identified or inferred in both experiments. 
    :rtype: Proteins
    """
    protein_one=exp1.get_proteins()
    protein_two=exp2.get_proteins()
    return list(protein_one.intersection(protein_two))

def compute_binary_distance(peptides: List[str], dist_func:Callable)->np.ndarray:
    """compare the distance between every pair of peptides in a collection of peptides. 
    
    :param peptides: a collection of peptides.
    :type peptides: List[str]
    :param dist_func: a function to compute the distance between each pair of peptides. 
    :type dist_func: Callable
    :raises RuntimeError: make sure that the dist_function is suitable with respect to the input peptides. For example, peptides which might have different lengths.
    :return: the distance between each pair of peptides in the provided list of peptides
    :rtype: np.ndarray
    """
    num_peptides=len(peptides)
    distance_matrix=np.zeros(shape=(num_peptides,num_peptides))
    # compute the pair wise distance 
    for raw_idx in range(num_peptides):
        for col_idx in range(num_peptides):
            try:
                distance_matrix[raw_idx,col_idx]=dist_func(peptides[raw_idx],peptides[col_idx])
            except Exception as exp:
                raise RuntimeError(f'While computing the pairwise distance between {peptides[raw_idx]} and {peptides[col_idx]} The following error was observed: {exp}')
    # return the results 
    return distance_matrix

def get_sequence_motif(peptides:Peptides, 
                       temp_dir: str ="./TEMP_DIR", verbose: bool = False, 
                       meme_params:Dict[str,str]={}
                       )->None:
    """compute the sequences motif from a collection of peptide sequences using meme software.
    
    :param peptides: a list of string containing the peptide sequences 
    :type peptides: Peptides
    :param temp_dir: the temp directory to write temp-files to it, defaults to "./TEMP_DIR"
    :type temp_dir: str, optional
    :param verbose: whether or not to print the output of the motif discovery tool to the stdout, defaults to False
    :type verbose: bool, optional
    :param meme_params: a dict object that contains meme controlling parameters, defaults to {}
    :type meme_params: Dict[str,str], optional
    :raises FileNotFoundError: incase meme is not installed or could not be found in the path!
    :raises ValueError: Incase the peptides have different length! 
    """
    # check the meme is installed 
    if not memeIF.is_meme_callable():
        raise FileNotFoundError(f"meme is either not installed or not part of the PATH ==> {os.environ['PATH']}")
    # check the temp directory
    try:
        os.mkdir(temp_dir)
    except FileExistsError: 
        pass 
    # check that amino acids have the same length 
    if len(set([len(pep) for pep in peptides])) !=1:
        raise ValueError('The provided peptides MUST have the same length!')
    # check that only the 20 classical amino acids are in the motif object 
    curated_sequences: List[str] =  [check_peptide_made_of_std_20_aa(pep) for pep in peptides]
    curated_sequences = [elem for elem in curated_sequences if elem !='']
    # write the sequences 
    outfile_seq_file=os.path.join(temp_dir,'TEMP_FASTA_SEQ.fasta')
    out_func.write_auto_named_peptide_to_fasta(peptides,outfile_seq_file)
    # call meme to compute the motif(s)
    meme_results_dir=os.path.join(temp_dir,'TEMP_MEME_RES')
    memeIF.call_meme(outfile_seq_file,
                output_dir=meme_results_dir, 
                verbose=verbose, **meme_params)
    print(f"MEME has finished execution. Results can be found at: {meme_results_dir}")
    
def download_structure_file(pdb_id:str)->None:
    """Download PDB/mmCIF file with a user provided identifer from PDB using BioPython library 

    :param pdb_id: the protein id in protein data bank 
    :type pdb_id: str
    """
    pdb_list=PDBList()
    pdb_list.retrieve_pdb_file(pdb_id)
    return 

def compute_expression_correlation(exp1:Experiment,exp2:Experiment)->float:
    """compute the correlation in the gene expression between two experiments by constructing a union
    of all the proteins expressed in the first and second experiments, extract the gene expression 
    of these genes and then compute the correlation using SciPy stat module. 
    
    :param exp1: The first experimental object 
    :type exp1: Experiment
    :param exp2: The second experimental object 
    :type exp2: Experiment
    :return: the correlation in gene expression of the proteins inferred in the provided pair of experiment
    :rtype: float
    """
    # get the expression tables 
    protein_exp1: Set[str] = set(exp1.get_proteins())
    protein_exp2: Set[str] = set(exp2.get_proteins())
    unique_proteins = list(protein_exp1.union(protein_exp2))
    # get the gene id 
    prot2Ense: pd.DataFrame = map_from_uniprot_gene(unique_proteins)
    # allocate lists to hold the results 
    gene_expression_exp1: List[float] = []
    gene_expression_exp2: List[float] = [] 
    # get the expression from experiment one 
    for prot in unique_proteins:
        temp_df: pd.DataFrame = prot2Ense.loc[prot2Ense.iloc[:,0]==prot]
        if temp_df.shape[0] == 1: # we got only one match 
            gene_id: str = temp_df['Gene-ID'].tolist()[0]
            try: 
                gene_expression_exp1.append(exp1.get_tissue().get_expression_profile().get_gene_id_expression(gene_id=gene_id))
            except KeyError:
                gene_expression_exp1.append(-1)
        else:
            temp_gene_expression: List[float] = []
            for gene in temp_df.iloc[:,1].tolist():
                try: 
                    temp_gene_expression.append(
                        exp1.get_tissue().get_expression_profile().get_gene_id_expression(gene_id=gene))
                except KeyError: 
                    temp_gene_expression.append(-1)
            # filter the temp_genes for default value 
            temp_gene_process: List[str] = [elem for elem in temp_gene_expression if elem != -1]
            # add the gene expression as the average if all values have been filtered 
            if len(temp_gene_process) ==0:
                gene_expression_exp1.append(-1)
            else: 
                gene_expression_exp1.append(np.mean(temp_gene_process))
    # ge the expression from exp2: 
    for prot in unique_proteins:
        temp_df: pd.DataFrame = prot2Ense.loc[prot2Ense.iloc[:,0]==prot]
        if temp_df.shape[0] == 1: # we got only one match 
            gene_id: str = temp_df['Gene-ID'].tolist()[0]
            try: 
                gene_expression_exp2.append(exp2.get_tissue().get_expression_profile().get_gene_id_expression(gene_id=gene_id))
            except KeyError:
                gene_expression_exp2.append(-1)
        else:
            temp_gene_expression: List[float] = []
            for gene in temp_df.iloc[:,1].tolist():
                try: 
                    temp_gene_expression.append(
                        exp2.get_tissue().get_expression_profile().get_gene_id_expression(gene_id=gene))
                except KeyError: 
                    temp_gene_expression.append(-1)
            # filter the temp_genes for default value 
            temp_gene_process: List[str] = [elem for elem in temp_gene_expression if elem != -1]
            # add the gene expression as the average 
            if len(temp_gene_process) ==0:
                gene_expression_exp2.append(-1)
            else: 
                gene_expression_exp2.append(np.mean(temp_gene_process))
    # compute construct a dataframe 
    temp_paired_exp_df: pd.DataFrame = pd.DataFrame({
        'exp2': gene_expression_exp1, 
        'exp1': gene_expression_exp2
    })
    # filter the un-mapped from exp1
    temp_paired_exp_df= temp_paired_exp_df.loc[temp_paired_exp_df.iloc[:,0]!=-1,]
    # filter the unmapped from exp2 
    temp_paired_exp_df= temp_paired_exp_df.loc[temp_paired_exp_df.iloc[:,1]!=-1,]
    
    # compute the correlation 
    return pearsonr(temp_paired_exp_df.iloc[:,0],temp_paired_exp_df.iloc[:,1])[0]

def compute_change_in_protein_representation(mapped_prot_cond1: np.ndarray, 
    mapped_prot_cond2: np.ndarray)->float:
    """Compute the change in the protein representation between two conditions, by computing 
    the differences in the area under the curve, AUC.

    :param mapped_prot_cond1: a mapped protein instance containing the protein coverage in the first condition
    :type mapped_prot_cond1: np.ndarray
    :param mapped_prot_cond2: a mapped protein instance containing the protein coverage in the second condition  
    :type mapped_prot_cond2: np.ndarray
    :raises ValueError: if the provided pair of proteins is of different length 
    :return: the difference in the area under the coverage curve between the two experiments. 
    :rtype: float
    """
    # un-roll the arrays to 1D arrays.
    if len(mapped_prot_cond1.shape)==2:
        mapped_prot_cond1=mapped_prot_cond1.reshape(-1)
    
    if len(mapped_prot_cond2.shape)==2:
        mapped_prot_cond2=mapped_prot_cond2.reshape(-1)
    
    # assert that the protein have same length 
    if len(mapped_prot_cond1)!=len(mapped_prot_cond2):
        raise ValueError(f'The provided proteins are of different length, found: {len(mapped_prot_cond1)} and {len(mapped_prot_cond1)}') 
    
    # compute the AUC 
    sum_array_one: float = np.sum(mapped_prot_cond1)
    sum_array_two: float = np.sum(mapped_prot_cond2)
    # compute the differences 
    difference: float = np.abs(sum_array_one-sum_array_two)
    # return the results 
    return difference

def compute_difference_in_representation(mapped_prot_cond1: np.ndarray, 
    mapped_prot_cond2: np.ndarray) -> np.ndarray:
    """return the difference in the representation of a protein between two conditions
    by substracting the coverage of the first protein from the second protein.

    :param mapped_prot_cond1: a mapped protein instance containing the protein coverage in the first condition
    :type mapped_prot_cond1: np.ndarray
    :param mapped_prot_cond2: a mapped protein instance containing the protein coverage in the second condition  
    :type mapped_prot_cond2: np.ndarray
    :return: an array that shows the difference in coverage between the two proteins at each amino acid position. 
    :rtype: np.ndarray
    """
     # un-roll the arrays to 1D arrays.
    if len(mapped_prot_cond1.shape)==2:
        mapped_prot_cond1=mapped_prot_cond1.reshape(-1)
    
    if len(mapped_prot_cond2.shape)==2:
        mapped_prot_cond2=mapped_prot_cond2.reshape(-1)
    
    # assert that the protein have same length 
    if len(mapped_prot_cond1)!=len(mapped_prot_cond2):
        raise ValueError(f'The provided proteins are of different length, found: {len(mapped_prot_cond1)} and {len(mapped_prot_cond1)}') 
    return mapped_prot_cond1-mapped_prot_cond2

def get_PTMs_modifications_positions(protein_feature: Features) ->List[int] : 
    r""" return a list of integers containing the position of PTMs in the protein.

    :param protein_feature: A protein feature instance containing all protein features 
    :type protein_feature: Features
    :return: A list that contains the position of the generic modifications in \
            the protein. If no modifications is known the function returns an empty list.
    :rtype: List[int]
    """
    containing=[]
    modification_dict=protein_feature.get_PTMs_modifications()
    if modification_dict==None: 
        return containing
    for key in modification_dict.keys():
            containing.append(modification_dict[key]["startIdx"])
    return containing

 
def get_PTMs_glycosylation_positions(protein_feature: Features) ->List[int]:
    r""" return a list of integers containing glycosylation position in the protein.

    :param protein_feature: a protein feature instance containing all protein features 
    :type protein_feature: Features
    :return: a list that contains the position of the generic glycosylation in \
            the protein. If no glycosylation site(s) are/is known in the protein \
            the function returns an empty list.
    :rtype: List[int]
    """
    glyco_positions=[]
    glyco_dict=protein_feature.get_PTMs_glycosylation()
    if glyco_dict==None: 
        return glyco_positions
    for key in glyco_dict.keys():
        glyco_positions.append(glyco_dict[key]["startIdx"])
    return glyco_positions 


def get_PTMs_disuldfide_bonds(protein_feature)->List[int]:
    """  return a list of integers containing dissulfide bound position in the protein.

    :param protein_feature: a protein feature instance containing all protein features 
    :type protein_feature: Features
    :return: A list that contains the position of the known disulfide bonds in \
            the protein. If no disulfide bond(s) is/are known in the protein \
            the function returns an empty list.
    :rtype: List[int]
    """
    disulfide_positions=[]
    disulfide_dict=protein_feature.get_disulfide_bonds()
    if disulfide_dict==None:
         return disulfide_positions
    for key in disulfide_dict.keys():
        disulfide_positions.append(disulfide_dict[key]["startIdx"])
    return disulfide_positions

def get_sequence_variants_positions(protein_feature)->List[int]:
    """return a list of integers containing known sequence variants positions in the protein.

    :param protein_feature: a protein feature instance containing all protein features 
    :type protein_feature: Features
    :return: A list that contains the position of the known sequence variants in 
            the protein. If no sequence variant/variants that is/are known in the protein,\
            the function returns an empty list.
    :rtype: List[int]
    """
    seqVar_positions=[]
    seqVar_dict=protein_feature.get_sequence_variants()
    if seqVar_dict==None:
        return seqVar_positions
    for key in seqVar_dict.keys():
        seqVar_positions.append(seqVar_dict[key]["startIdx"])
    return seqVar_positions

def get_chain_positions(protein_feature)->List[List[int]]:
    """ return a list of lists containing the start and end position of known chains in the protein 

    :param protein_feature: a protein feature instance containing all protein features 
    :type protein_feature: Features
    :return:  a list of list that contains the position of the known chain/chains \
            in the protein. each list has two elements which are the start and \
            the end position of the position lists. \
            If no chains are known in  the protein, \
            the function returns an empty list. \
    :rtype: List[List[int]]
    """
    chain_positions=[]
    chains_dict=protein_feature.get_chains()
    if chains_dict==None: 
        return chain_positions
    for key in chains_dict.keys():
            chain_positions.append([chains_dict[key]["startIdx"],
                                    chains_dict[key]["endIdx"]])
    return chain_positions


def get_domains_positions(protein_feature)->List[List[int]]:
    """ return a list of lists containing the start and end position of known domains in the protein.  

    :param protein_feature: a protein feature instance containing all protein features 
    :type protein_feature: Features
    :return: a list of list that contains the position of the known domain/domains \
            in the protein. each list has two elements which are the start and \
            the end position of the domain. \
            If there are no domain/domains that is/are known in the protein, \
            the function returns an empty list. \
    :rtype: List[List[int]]
    """
    domains_positions=[]
    domains_dict=protein_feature.get_domains()
    if domains_dict==None: 
        return domains_positions
    for key in domains_dict.keys():
        domains_positions.append([domains_dict[key]["startIdx"],
                                    domains_dict[key]["endIdx"]])
    return domains_positions

def get_splice_variants_positions(protein_feature)->List[List[int]]:
    """ return a list of lists containing the start and end position of known splice variants in the protein.  

    :param protein_feature: a protein feature instance containing all protein features 
    :type protein_feature: Features
    :return:  a list of list that contains the position of the known splice variant/variants \ 
            in the protein. each list has two elements which are the start and \
            the end position of the splice variants. \
            If no splice variant/variants that is/are known in the protein, \
            the function returns an empty list. \
    :rtype: List[List[int]]
    """
    splice_variant_positions=[]
    variants=protein_feature.get_splice_variants()
    if variants==None: 
        return splice_variant_positions
    for key in variants.keys():
        splice_variant_positions.append([variants[key]["startIdx"],
                                    variants[key]["endIdx"]])
    return splice_variant_positions

def compute_ic_distance_protein(protein_id:str, experiment_set,
                                mode="restrictive") -> pd.DataFrame: 
    """compute the immunopeptidomic coverage distance between a group of experiments \
        in the experiment set using one protein defined in protein_id. 

    :param protein_id: the uniprot or the identifier of the protein in Experiment objects
    :type protein_id: str
    :param experiment_set: An experiment set object containing all the experiments 
    :type experiment_set: ExperimentSet
    :param mode: mode of calculations, if restrictive, the protein defined by protein_id MUST\
         be present in every experiment. If is not defined an error will be through. However,\
         incase the mode is "permissive" the absent protein will be treated as an array of zeros.\
         it defaults to "restrictive".
    :type mode: str, optional
    :return: a square distance matrix contain the differences in immunopeptidomics coverage between each pair of experiments\
         in the set of experiments. 
    :rtype: pd.DataFrame
    """
    # extract the mapping array: 
    exps=experiment_set.get_experiments()
    mapped_arrays=dict()
    ## loop over each experiment and extract the experiment 
    for exp_id in exps.keys(): 
        try: 
            mapped_arrays[exp_id]=exps[exp].get_mapped_protein(protein_id)
        except KeyError as exp: 
            if mode=="permissive": 
                mapped_arrays[exp_id]=-1
            else: 
                raise KeyError(f"protein id: {protein_id} is not defined in experiment: {exp_id}")
    ## compute the pair wise distance 
    ic_dist: np.ndarray = np.zeros((len(mapped_arrays),len(mapped_arrays)))
    ## get the names of the arrays  
    row_names=col_names=list(mapped_arrays.keys())
    ## loop over the array element to fill 
    for row_idx in range(len(row_names)): 
        for col_idx in range(len(col_names)): 
            # handling the 4 possible cases that can be encountered during the execution 
            if isinstance(mapped_arrays[row_names[row_idx]], int): 
                if isinstance(mapped_arrays[row_names[col_idx]], int):
                    ic_dist[row_idx,col_idx]=0
                else: 
                    ic_dist[row_idx,col_idx]=np.sum(np.abs(mapped_arrays[row_names[col_idx]]))
            else: 
                if isinstance(mapped_arrays[row_names[col_idx]], int):
                    ic_dist[row_idx,col_idx]=np.sum(np.abs(mapped_arrays[row_names[row_idx]]))
                else: 
                    ic_dist[row_idx,col_idx]=np.sum(np.abs(mapped_arrays[row_names[row_idx]]-mapped_arrays[row_names[col_idx]]))
    ## construct a data frame out of the results 
    results_df: pd.DataFrame = pd.DataFrame(mapped_arrays)   
    results_df.columns=col_names
    results_df.index=row_names
    ## return the results 
    return results_df

def compute_ic_distance_experiments(experiment_set,mode="restrictive")->pd.DataFrame:
    """compute the immunopeptidomic coverage distance between a group of experiments \
        using all proteins in the intersection or the union, depending on the mode, of the set\
        of proteins defined in the set. 

    :param experiment_set: An experiment set object containing all the experiments 
    :type experiment_set: ExperimentSet
    :param mode: mode of calculations, if restrictive the proteins defined by protein_id MUST\
        be defined in every experiment, if it is not defined and error will be through. However,\
        incase it is "permissive" the absent protein will be treated as an array of zeros.\
        it defaults to "restrictive".
    :type mode: str, optional
    :return: A square distance matrix containing the avarage difference in the immunopeptidomic coverage\
        between each pair of experiments in the experiment set.  
    :rtype: pd.DataFrame
    """
    ## GETTING the set of proteins depending on the mode
    if mode=="restrictive": 
        protin_set=experiment_set.get_proteins_present_in_all()
    else: 
        protein_set=[]
        for exp_id in experiment_set.get_experiments().keys(): 
            protein_set.extend(experiment_set[exp_id].get_proteins())
    protin_set=list(set(protin_set))
    ## getting the dict of experiments 
    exps=experiment_set.get_experiments()
    ## computing the distance tensor Tensor 
    distance_tensor = np.zeros((len(exps),len(exps), len(protin_set)))
    ## looping over all the elements in the tensor 
    row_names=col_names= list(exps.keys())
    ## FILL the tensor  
    # ----------------
    for row_idx in range(len(row_names)):
        for col_idx in range(len(col_names)): 
            for protein_idx in range(len(protin_set)):
                ## obtain the mapped array of the rows
                try: 
                    mapped_array_er_pp=exps[row_names[row_idx]].get_mapped_protein(protin_set[protein_idx])
                except KeyError: 
                    mapped_array_er_pp=-1
                ## obtain the mapped array of the columns 
                try: 
                    mapped_array_ec_pp= exps[col_names[col_idx]].get_mapped_protein(protin_set[protein_idx]) 
                except KeyError: 
                    mapped_array_er_pp=-1
                ## fill the tensor using the mapped array information 
                if isinstance(mapped_array_er_pp, int): # i.e. set to -1 --> not present   
                    if isinstance(mapped_array_ec_pp, int): # i.e. set to -1 --> not present  
                        distance_tensor[row_idx,col_idx,protein_idx]=0
                    else: 
                        distance_tensor[row_idx,col_idx,protein_idx]=np.sum(np.abs(mapped_array_ec_pp))
                else: 
                    if isinstance(mapped_array_ec_pp, int):
                        distance_tensor[row_idx,col_idx,protein_idx]=np.sum(np.abs(mapped_array_er_pp))
                    else:
                        distance_tensor[row_idx,col_idx,protein_idx]=np.sum(np.abs(mapped_array_er_pp-mapped_array_ec_pp)) 
    ## COMPUTE the mean across the protein set: 
    distance_matrix= np.mean(distance_tensor,-1) 
    ## CONSTRUCT A DF out of the results 
    results_df=pd.DataFrame(distance_matrix)
    ## set the names of the dataframe 
    results_df.columns=col_names
    results_df.index=row_names
    ## return the results 
    return results_df






#!/usr/bin/env python 
"""Parses the XML scheme of a uniprot protein and provides a python API  
for quering and accessing the results 
"""
# load the modules 
from __future__ import annotations
import numpy as np 
from Bio import SeqIO
import urllib
import os 
from typing import List, Tuple, Dict, Union
class Features:
    r"""The class provides a template for the features associated with a protein. \
    The following features are associated with the protein \ 
    #signal peptide: dict \
        The range of the signal peptides, if the protein has no signal, for example, a globular \
        cytologic protein. None is used as a default, placeholder \
        value. \ 
    #chains:dict \ 
        the chains making up the mature protein, the protein should at least have one chain. \ 
    #domain: dict\ 
        the known domains in the protein, if no domain is defined, None is used. \ 
    #modification sites: nested dict \ 
        that contains information about the PTM sites, glycosylation site and disulfide bonds. \ 
    #sequence variances: dict \ 
        which contains information about the sequence variants of a protein structure. \ 
    #split variance: dict \ 
        which contain known splice variants \ 
    ** Notes: Although disulfide bond is not a PTMs, it is being treated as a \ 
    one here to simplify the workflow. \ 
    """
    def __init__(self,uniprot_id:str, temp_dir: str = None)->Features:
        """Download the features associated with a protein from uniprot and then parse the results using SeqIO to extract the features
        
        :param uniprot_id: Uniprot id to download its XML scheme from Uniprot 
        :type uniprot_id: str 
        :param temp_dir: a temporary directory to download the XML scheme to it, if not provided files are download to the current working directory, defaults to None.
        :type temp_dir: str
        """
        # download the files 
        if temp_dir is None: 
            temp_dir="." 
        # define the file path 
        file_path: str = os.path.join(temp_dir,f"{uniprot_id}.xml")
        try: 
            urllib.request.urlretrieve(f"https://www.uniprot.org/uniprot/{uniprot_id}.xml", file_path) 
        except urllib.error.HTTPError: 
            raise IOError("Downloading the provided fail, Check your provided uniprot id")
        # read the sequence object 
        record: SeqIO.SeqRecord = SeqIO.read(file_path,"uniprot-xml")
        # fill the provided record 
        self.extracted_features=dict()
        # parse the features in the record.feature object
        for feature in record.features:
            # extract sequences signal peptide
            if feature.qualifiers["type"]=="signal peptide":
                # Try to extract the start and end position for the chain
                # if this could no be extracted None is used as a default value
                try:
                    start_index=int(feature.location.start)
                except(TypeError):
                    start_index=None
                try:
                    end_index=int(feature.location.end)
                except(TypeError):
                    end_index=None
                self.extracted_features["SignalPeptide"]={
                    "startIdx":start_index,
                    "endIdx":end_index
                    }
            # extract the chain information
            elif feature.qualifiers["type"]=="chain":
                if "Chains" in self.extracted_features.keys():
                    chainIdx=len(self.extracted_features["Chains"]) # get the chain index
                    chainName="chain_number_"+str(chainIdx)
                    # Try to extract the start and end position for the chain
                    # if this colud no be extracted None is used as a default value
                    try:
                        start_index=int(feature.location.start)
                    except(TypeError):
                        start_index=None
                    
                    try:
                        end_index=int(feature.location.end)
                    except(TypeError):
                        end_index=None
                    self.extracted_features["Chains"][chainName]={
                        "chainId":feature.id,
                        "startIdx":start_index,
                        "endIdx":end_index
                        }
                else:
                    
                    try:
                        start_index=int(feature.location.start)
                    except(TypeError):
                        start_index=None
                    
                    try:
                        end_index=int(feature.location.end)
                    except(TypeError):
                        end_index=None
                        
                    self.extracted_features["Chains"]={
                        "chain_number_0":{
                        "chainId":feature.id,
                        "startIdx":start_index,
                        "endIdx":end_index
                            }}
            # extract the domain information
            elif feature.qualifiers["type"]=="domain":
                if "Domains" in self.extracted_features.keys():
                    self.extracted_features["Domains"][
                    feature.qualifiers["description"]]={
                        "startIdx":int(feature.location.start),
                        "endIdx":int(feature.location.end)}
                else: 
                    self.extracted_features["Domains"]={
                        feature.qualifiers["description"]:{
                        "startIdx":int(feature.location.start),
                        "endIdx":int(feature.location.end)
                            }
                        }
            # extract the topological information:
            elif feature.qualifiers["type"]=="modified residue":
                if "PTMs" in self.extracted_features.keys():
                    if "Modifications" in self.extracted_features["PTMs"].keys():
                        modificationIdx=len(self.extracted_features["PTMs"][
                            "Modifications"])
                        modificationName="SeqModification_num_"+str(modificationIdx)
                        self.extracted_features["PTMs"]["Modifications"][
                            modificationName]={
                                "Name":feature.qualifiers["description"],
                                "startIdx":int(feature.location.start),
                                "endIdx":int(feature.location.end)
                                }
                    else:
                        self.extracted_features["PTMs"]["Modifications"]={}
                        self.extracted_features["PTMs"]["Modifications"][
                            "SeqModification_num_0"]={
                            "Name":feature.qualifiers["description"],
                            "startIdx":int(feature.location.start),
                            "endIdx":int(feature.location.end)
                            }
                else: 
                    self.extracted_features["PTMs"]={}
                    self.extracted_features["PTMs"]["Modifications"]={}
                    self.extracted_features["PTMs"]["Modifications"][
                        "SeqModification_num_0"]={
                            "Name":feature.qualifiers["description"],
                            "startIdx":int(feature.location.start),
                            "endIdx":int(feature.location.end)
                            }
            # extract and add glycosylation sites to the features class
            elif feature.qualifiers["type"]=="glycosylation site":
                if "PTMs" in self.extracted_features.keys():
                    if "GlycoSite" in self.extracted_features["PTMs"].keys():
                        glycositeIdx=len(self.extracted_features["PTMs"]["GlycoSite"])
                        glyco_site_name="Glyco_Site_number_"+str(glycositeIdx)
                        self.extracted_features["PTMs"]["GlycoSite"][
                            glyco_site_name]={
                                "Name":feature.qualifiers["description"],
                                "startIdx":int(feature.location.start),
                                "endIdx":int(feature.location.end)
                                }
                    else:
                        self.extracted_features["PTMs"]["GlycoSite"]={}
                        self.extracted_features["PTMs"]["GlycoSite"][
                            "Glyco_Site_number_0"]={
                                "Name":feature.qualifiers["description"],
                                "startIdx":int(feature.location.start),
                                "endIdx":int(feature.location.end)
                                }
                else:
                    self.extracted_features["PTMs"]={}
                    self.extracted_features["PTMs"]["GlycoSite"]={}
                    self.extracted_features["PTMs"]["GlycoSite"][
                            "Glyco_Site_number_0"]={
                                "Name":feature.qualifiers["description"],
                                "startIdx":int(feature.location.start),
                                "endIdx":int(feature.location.end)
                                }
            # extract and add disulfide site to the feature class
            elif feature.qualifiers["type"]=="disulfide bond":
                if "PTMs" in self.extracted_features.keys():
                    if "DisulfideBond" in self.extracted_features["PTMs"].keys():
                        disulfideBondIdx=len(self.extracted_features["PTMs"]["DisulfideBond"])
                        disulfide_site_name="disulfide_site_number_"+str(disulfideBondIdx)
                        # try to extract the start and the end position.
                        try:
                            start_index=int(feature.location.start)
                        except(TypeError):
                            start_index=None
                        try:
                            end_index=int(feature.location.end)
                        except(TypeError):
                            end_index=None
                        self.extracted_features["PTMs"]["DisulfideBond"][
                            disulfide_site_name]={
                                "Name":feature.qualifiers["type"],
                                "startIdx":start_index,
                                "endIdx":end_index
                                }
                    else:
                        try:
                            start_index=int(feature.location.start)
                        except(TypeError):
                            start_index=None
                        try:
                            end_index=int(feature.location.end)
                        except(TypeError):
                            end_index=None
                        self.extracted_features["PTMs"]["DisulfideBond"]={}
                        self.extracted_features["PTMs"]["DisulfideBond"][
                            "disulfide_site_number_0"]={
                                "Name":feature.qualifiers["type"],
                                "startIdx":start_index,
                                "endIdx":end_index
                                }
                else:
                    self.extracted_features["PTMs"]={}
                    self.extracted_features["PTMs"]["DisulfideBond"]={}
                    self.extracted_features["PTMs"]["DisulfideBond"][
                            "disulfide_site_number_0"]={
                                "Name":feature.qualifiers["type"],
                                "startIdx":int(feature.location.start),
                                "endIdx":int(feature.location.end)
                                }
            # extract sequence variant from the protein sequence:
            elif feature.qualifiers["type"]=="sequence variant":
                if "SeqVar" in self.extracted_features.keys():
                    varient_index=len( self.extracted_features["SeqVar"])
                    varient_name="Sequence_varient_number_"+str(varient_index)
                    # check if the feature has a description entrey
                    if "description" in feature.qualifiers.keys():
                        snp_id=feature.qualifiers["description"]
                    else:
                        snp_id=None
                    # check if the feature has a original entrey
                    if "original" in feature.qualifiers.keys():
                        original_amino_acid=feature.qualifiers["original"]
                    else:
                        original_amino_acid=None
                    # check if the feature has a varient entrey
                    if "variation" in feature.qualifiers.keys():
                        varient_amino_acid=feature.qualifiers["variation"]
                    else:
                        varient_amino_acid=None
                    # fill the entries
                    self.extracted_features["SeqVar"][varient_name]={
                        "VarientId":feature.qualifiers["id"],
                        "SNP_Id":snp_id,
                        "original":original_amino_acid,
                        "variation":varient_amino_acid,
                        "startIdx":int(feature.location.start),
                        "endIdx":int(feature.location.end)
                        }
                else:
                    self.extracted_features["SeqVar"]={}
                    if "description" in feature.qualifiers.keys():
                        snp_id=feature.qualifiers["description"]
                    else:
                        snp_id=None
                    # check if the feature has a description entrey
                    if "description" in feature.qualifiers.keys():
                        snp_id=feature.qualifiers["description"]
                    else:
                        snp_id=None
                    # check if the feature has a original entrey
                    if "original" in feature.qualifiers.keys():
                        original_amino_acid=feature.qualifiers["original"]
                    else:
                        original_amino_acid=None
                    # check if the feature has a varient entrey
                    if "variation" in feature.qualifiers.keys():
                        varient_amino_acid=feature.qualifiers["variation"]
                    else:
                        varient_amino_acid=None
                    self.extracted_features["SeqVar"][
                        "Sequence_varient_number_0"]={
                        "VarientId":feature.qualifiers["id"],
                        "SNP_Id":snp_id,
                        "original":original_amino_acid,
                        "variation":varient_amino_acid,
                        "startIdx":int(feature.location.start),
                        "endIdx":int(feature.location.end)
                        }
            # extract splice vaients from the protein sequences: 
            elif feature.qualifiers["type"]=="splice variant":
                if "SpliceVar" in self.extracted_features.keys():
                    SpliceVarIdx=len(self.extracted_features["SpliceVar"])
                    spliceVarient_name="splice_varient_number_"+str(SpliceVarIdx)
                    self.extracted_features["SpliceVar"][spliceVarient_name]={
                        "Name":feature.qualifiers["id"],
                        "Isoform":feature.qualifiers["description"],
                        "startIdx":int(feature.location.start),
                        "endIdx":int(feature.location.end)
                        }
                else: 
                    self.extracted_features["SpliceVar"]={}
                    self.extracted_features["SpliceVar"][
                        "splice_varient_number_0"]={
                        "Name":feature.qualifiers["id"],
                        "Isoform":feature.qualifiers["description"],
                        "startIdx":int(feature.location.start),
                        "endIdx":int(feature.location.end)
                        }
        # fill in the empty object with None:
        if "SignalPeptide" not in self.extracted_features.keys():
            self.extracted_features["SignalPeptide"]=None
        if "Chains" not in self.extracted_features.keys():
            self.extracted_features["Chains"]=None
        if "Domains" not in self.extracted_features.keys():
            self.extracted_features["Domains"]=None
        if "PTMs" not in self.extracted_features.keys():
            self.extracted_features["PTMs"]=None
        else:
            if "Modifications" not in self.extracted_features["PTMs"].keys():
                self.extracted_features["PTMs"]["Modifications"]=None 
            if "GlycoSite" not in self.extracted_features["PTMs"].keys():
                self.extracted_features["PTMs"]["GlycoSite"]=None 
            if "DisulfideBond" not in self.extracted_features["PTMs"].keys():
                self.extracted_features["PTMs"]["DisulfideBond"]=None 
        if "SeqVar" not in self.extracted_features.keys():
            self.extracted_features["SeqVar"]=None
        if "SpliceVar" not in self.extracted_features.keys():
            self.extracted_features["SpliceVar"]=None
            
    # accessor methods:
    def has_signal_peptide(self)->bool:
        """
        :return: True if the protein has a signal peptide and False other wise.
        :rtype: bool
        """
        if self.extracted_features["SignalPeptide"]==None:
            return False
        return True
        
    def get_signal_peptide_index(self)->Tuple[int,int]:
        """
        :return:  The Index of the signal-peptide in the protein, if not signal peptide is defined, it returns None
        :rtype: Tuple[int,int]
        """
        if self.extracted_features["SignalPeptide"]==None:
            return None,None
        startIdx=self.extracted_features["SignalPeptide"]["startIdx"]
        endIdx=self.extracted_features["SignalPeptide"]["endIdx"]
        return startIdx,endIdx
    
    def has_chains(self)->bool:
        """
        :return: True if the protein has/have chain/chains as feature and False otherwise.
        :rtype: [type]
        """
        if self.extracted_features["Chains"]==None:
            return False
        return True
    
    def get_number_chains(self) -> int:
        """
        :return: The number of chains in the protein. if no chain is defined it returns zero.
        :rtype: int
        """
        if not self.has_chains():
            return 0
        return len(self.extracted_features["Chains"])
    
    def get_chains(self)->Dict[Dict[str,Union[str,int]]]:
        """
        :return: A dictionary that contains the chains of the protein, if no chain is defined it return None
        :rtype: Dict[Dict[str,Union[str,int]]]
        """
        return self.extracted_features["Chains"]
    
    def has_domains(self)->bool: 
        """
        :return: True if the protein has a defined domain/domains, otherwise it return False.
        :rtype: bool 
        """
        if self.extracted_features["Domains"]==None:
            return False
        return True
    
    def get_number_domains(self)->int:
        """
        :return: The number of domains a protein has, if no domain is defined it returns zero.
        :rtype: int
        """
        if self.extracted_features["Domains"] ==None:
            return 0
        return len(self.extracted_features["Domains"])
    
    def get_domains(self)->Dict[str, Dict[str, int]]:
        """
        :return:  The domains defined in the protein sequence, if no domain is defined it returns None.
        :rtype: Dict[str, Dict[str, int]]
        """
        return self.extracted_features["Domains"]
    
    def has_PTMs(self)->bool:
        """
        :return:True if the protein has a PTMs and False other wise
        :rtype: bool
        """
        if self.extracted_features["PTMs"] ==None:
            return False
        return True

    def has_disulfide_bond(self)->bool:
        """
        :return: True is the protein has disulfide and False other wise
        :rtype: bool
        """
        if self.has_PTMs():
            if self.extracted_features["PTMs"]["DisulfideBond"]==None:
                return False
            else: 
                return True
        return False
    
    def has_glycosylation_site(self)->bool:
        """
        :return: True if the protein has a glycosylation site and False otherwise.
        :rtype: [type]
        """
        if self.has_PTMs():
            if self.extracted_features["PTMs"]["GlycoSite"] == None:
                return False
            else:
                return True
        return False
    
    def has_site_modifications(self)->bool:
        """
        :return: True if the protein has a modification site and False otherwise
        :rtype: bool
        """
        if self.has_PTMs():
            if self.extracted_features["PTMs"]["Modifications"] == None:
                return False
            else:
                return True
        return False
    
    def get_PTMs(self)->Dict[str,Dict[str,Dict[str,Union[str,int]]]]:
        """
        :return:  a nested dictionary that contains the PTMs found within the protein \
        the PTMs are classified into three main categories:

            1- Modifications: which is the generic case and contain information \
            about any sequence modification beside disulfide bonds and glycosylation.
            
            2- glycosylation: contains information about glycosylation sites
            
            3- DisulfideBond: contains information about disulfide bond

        :rtype: Dict[str,Dict[str,Dict[str,Union[str,int]]]]
        """
        return self.extracted_features["PTMs"]
    
    def get_PTMs_modifications(self)->Dict[str,Dict[str,Union[str,int]]]:
        """
        :return:  The generic modifications found on the protein. If the protein has no PTM, the function returns None.
        :rtype: Dict[str,Dict[str,Union[str,int]]]
        """
        return self.extracted_features["PTMs"]["Modifications"]
    
    def get_PTMs_glycosylation(self)->Dict[str,Dict[str,Union[str,int]]]:
        """
        :return: The glycosylation sites found on the protein. If the protein has no glycosylation sites, the function returns None.
        :rtype: [type]
        """
        return self.extracted_features["PTMs"]["GlycoSite"]
    
    def get_disulfide_bonds(self)->Dict[str,Dict[str,Union[str,int]]]:
        """
        :return: The disulfide sites found on the protein. If the protein has no disulfide sites, the function returns None
        :rtype: [type]
        """
        return self.extracted_features["PTMs"]["DisulfideBond"]
    
    def get_number_PTMs(self)->int:
        """
        :return:  The number of PTMs the sequence has, this include di-sulfide bonds. See Note1 for more details. \
        If the protein has no PTMs the function returns zero
        :rtype: int
        """
        if not self.has_PTMs():
            return 0
        else:
             number_of_PTMs=self.get_number_modifications()+self.get_number_glycosylation_sites()+self.get_number_disulfide_bonds()
             return number_of_PTMs
    
    def get_number_modifications(self)->int:
        """
        :return:  Returns the total number of generic modifications found on the protein. \
        if no modification is found it return 0
        :rtype: int
        """
        if not self.has_site_modifications():
            return 0
        else:
            return len(self.get_PTMs_modifications())
    
    def get_number_glycosylation_sites(self)->int:
        """
        :return: The number of glycosylation_sites the protein has, if the protein has no glycosylation sites, the function returns zero
        :rtype: int
        """
        if not self.has_glycosylation_site():
            return 0
        else:
            return len(self.get_PTMs_glycosylation())
    
    def get_number_disulfide_bonds(self)->int:
        """
        :return: The number of disulfide bonds the protein has, if the protein has no disulfide bonds, the function return zero.
        :rtype: int
        """
        if not self.has_disulfide_bond():
            return 0
        else:
            return len(self.get_disulfide_bonds())
        
    def has_sequence_variants(self) ->bool:
        """
        :return:  True if the protein has a sequence variants, and False otherwise.
        :rtype: bool
        """
        if self.extracted_features["SeqVar"] == None:
            return False
        else: 
            return True
    
    def get_sequence_variants(self) -> Dict[str,Dict[str,Union[str,int]]]:
        """
        :return: A dict object that contains all sequence variants within a protein, if the protein has no sequence variants the function returns None.
        :rtype: Dict[str,Dict[str,Union[str,int]]]
        """
        return self.extracted_features["SeqVar"]
    
    def get_number_sequence_variants(self)->int:
        """
        :return: The number of sequence variants the protein has, if the protein has no sequence varient, the function returns 0.
        :rtype: int
        """
        if not self.has_sequence_variants():
            return 0
        else:
            return len(self.get_sequence_variants())
    
    def has_splice_variants(self)->bool:
        """
        :return: True if the sequence has a splice variants and False otherwise.
        :rtype: bool
        """
        if self.extracted_features["SpliceVar"]==None:
            return False
        else:
            return True
    
    def get_splice_variants(self)->Dict[str,Dict[str,Union[str,int]]]:
        """
        :return: A dict object that contains the splice variants. If the protein has no splice variants the function returns None.
        :rtype: Dict[str,Dict[str,Union[str,int]]]
        """
        return self.extracted_features["SpliceVar"]
    
    def get_number_splice_variants(self)->int:
        """
        :return: The number of slice variants in the protein, if the protein has no splice variants, the function returns zero.
        :rtype: int
        """
        if not self.has_splice_variants():
            return 0
        else:
            return len(self.get_splice_variants())
    
    def summary(self)->Dict[str,Union[str,int]]:
        """
        :return: The function return a dict object that summarizes the features of the protein.
        :rtype: Dict[str,Union[str,int]]
        """
        summary=dict()
        summary["has_signal_peptide"]=self.has_signal_peptide()
        summary["number_of_chains"]=self.get_number_chains()
        summary["number_of_domains"]=self.get_number_domains()
        summary["number_of_PTMs"]=self.get_number_PTMs()
        summary["number_of_modifications"]=self.get_number_modifications()
        summary["number_of_glycosylation_sites"]=self.get_number_glycosylation_sites()
        summary["number_of_disulfide_bonds"]=self.get_number_disulfide_bonds()
        summary["number_of_sequence_varients"]=self.get_number_sequence_variants()
        summary["number_of_splice_varients"]=self.get_number_splice_variants()
        return summary
    
    def __str__(self)->str:
        """
        The string representation of the class
        Returns
        -------
        the string representation of the class
        """
        summary=self.summary()
        string_rep=""" a protein feature class with: {}  chains, {} domains.  {} PTMs, {} sequence varients and {} splice varients""".format(
            summary["number_of_chains"],summary["number_of_domains"],
            summary["number_of_PTMs"],summary["number_of_sequence_varients"]   , summary["number_of_splice_varients"])
        return string_rep

    def __repr__(self)->str: 
        """
        :return: A formated print statement for the class 
        :rtype: str
        """
        return str(self)
    


#!/usr/bin/env Python 
"""Utility functions that are used through the library 
"""
# load the models 
from Bio import SeqIO
from IPTK.Utils.Types import FastaSet
import numpy as np
import pandas as pd
import os
import matplotlib.pyplot as plt
import matplotlib 
import string
import random
import pickle 
import urllib
from typing import List, Dict
# define the functions 
def pad_mapped_proteins(list_array: List[np.ndarray],
                pre_pad:bool =True, padding_char: int =-1)->np.ndarray:
    """ Pad the provided list of array into a 2D tensor of shape number of arrays by maxlength. 

    :param list_array: A list of NumPy arrays where each array is a mapped_protein array, \
    the expected shape of these arrays is 1 by protein length.
    :type list_array: List[np.ndarray]
    :param pre_pad: pre or post padding of shorter array in the list_array. Defaults to True, which mean prepadding
    :type pre_pad: bool, optional
    :param padding_char: The padding char, defaults to -1
    :type padding_char: int, optional
    :return: A 2D tensor of shape number of arrays by maxlength. 
    :rtype: np.ndarray
    """
    # reshape the resulting arrays
    for idx in range(len(list_array)):
        list_array[idx].reshape(1,-1) # reshape te array into shape 1* number of elements 
    # getting the max 
    max_len: int =max([elem.shape[1] for elem in list_array])
    # compute the padding distance 
    paddling_lens: List[int]=[max_len-elem.shape[1] for elem in list_array]
    # making arrays to hold padding_arrays
    padding_arrays: List[np.ndarray]= []
    # generate the padding distance 
    for idx in range(len(list_array)):
        padding_arrays.append(np.array([padding_char]*paddling_lens[idx]).reshape(1,-1))
    # allocate a list to hold the results 
    resulting_arrays: List[np.ndarray] = []
    # fuse the arrays 
    for idx in range(len(list_array)):
        if pre_pad:
            resulting_arrays.append(
                np.concatenate([padding_arrays[idx], list_array[idx]],axis=1)
            )
        else: 
            resulting_arrays.append(
                np.concatenate([list_array[idx], padding_arrays[idx] ],axis=1)
            )
    # concat the results array 
    results_array: np.ndarray = np.concatenate(resulting_arrays,axis=0)
    return results_array

def generate_random_name(name_length: int)->str:
    """
    :param name_length: Generate a random ASCII based string
    :type name_length: int
    :return: [description]
    :rtype: str
    """
    chars=[char for char in string.ascii_uppercase] # get ASCII upper chars 
    chars.extend([str(elem) for elem in list(range(10))]) # extend the vocab 
    return ''.join([random.choice(chars) for _ in range(name_length)]) # generate the name 

def append_to_calling_string(param:str, def_value, cur_val, calling_string:str, is_flag: bool = False)->str:
    """ help function that take a calling string, a parameter, a default value and current value \
    if the parameter does not equal its default value the function append the parameter with its current \ 
    value to the calling string adding a space before the calling_string. 

    :param param: The name of the parameter that will be append to the calling string 
    :type param: str
    :param def_value: The default value for the parameter 
    :type def_value: [type]
    :param cur_val: The current value for the parameter
    :type cur_val: [type]
    :param calling_string: The calling string in which the parameter and the current value might be appended to it 
    :type calling_string: str
    :param is_flag: If the parameter is a control flag, i.e. a boolean switch, it append the parameter to the calling string without associating a value to it , defaults to False
    :type is_flag: bool, optional
    :return: the updated version of the calling string 
    :rtype: str
    """
    if is_flag: 
        if def_value != cur_val:
            calling_string += ' -'+param
        return calling_string
    if def_value != cur_val: 
        calling_string+=" -"+param+" "+str(cur_val)
    return calling_string

def generate_random_protein_mapping(protein_len: int , max_coverage: int) -> np.ndarray:
    """Generate a NumPy array with shape of 1 by protein_len where the elements in the array 
    is a random integer between zero &  max_coverage. 

    :param protein_len: The length of the protein 
    :type protein_len: int
    :param max_coverage: The maximum peptide coverage at each position 
    :type max_coverage: int
    :return: a NumPy array containing a simulated protein coverage 
    :rtype: np.ndarray
    """
    return np.random.randint(low=0, high= max_coverage, size=(protein_len,))

def generate_color_scale(color_ranges: int )-> matplotlib.colors.LinearSegmentedColormap: 
    """generate a color gradient with number of steps equal to color_ranges -1 
    
    :param color_ranges:  the number of colors in the range
    :type color_ranges: int
    :return: A color gradient palette
    :rtype: matplotlib.colors.LinearSegmentedColormap
    """
    return plt.cm.get_cmap('hsv',color_ranges)

def simulate_protein_representation(num_conditions : int , protein_len: int ,
         protein_coverage: int )->Dict[str, np.ndarray]:
    """ Simulate protein peptide coverage under-different conditions
    
    :param num_conditions: The number of condition to simulate 
    :type num_conditions: [type]
    :param protein_len: The length of the protein  
    :type protein_len: [type]
    :param protein_coverage: The maximum protein coverage 
    :type protein_coverage: [type]
    :return: a dict of length num_conditions containing the condition index and a simulated protein array   
    :rtype: Dict[str, np.ndarray]
    """
    sim_res=dict()
    color_gradient=generate_color_scale(num_conditions)
    for idx in range(num_conditions): 
        sim_res['COND_'+str(idx)]={'name': 'COND_'+str(idx),
                                   'mapped_protein':generate_random_protein_mapping(protein_len,protein_coverage),
                                   'color':color_gradient(idx)
                                   }
    return sim_res

def simulate_protein_binary_represention(num_conditions: int, protein_length: int): 
    """
    :param num_conditions: The number of conditions to simulate 
    :type num_conditions: int
    :param protein_length: The Length of the protein  
    :type protein_length: int
    :return: A 2D matrix of shape protein_length by number of conditions, where each element can be either zero or 1.
    :rtype: np.ndarray
    """
    return np.random.randint(low=0, high=2, size=(protein_length,num_conditions)).astype(np.float64)
 
def save_3d_figure(outpath: str, fig2save: plt.Figure) ->None:
    """write a pickled version of the a 3D figure so it can be loaded later for more interactive analysis
    
    :param outpath: The output path of the writer function 
    :type outpath: str
    :param fig2save: The figure to save to the output file
    :type fig2save: plt.Figure
    :raises IOError: In case writing the file failed 
    """
    try: 
        with open(outpath,'wb') as writer_buf:
            pickle.dump(fig2save,writer_buf)
    except Exception as exp:
        raise IOError(f'While writing the figure to the provided path the following error was encountered: {exp}')

def load_3d_figure(file_path: str) ->plt.Figure: 
    """
    :param file_path: Load a pickled 3D figure from the provided path 
    :type file_path: str
    :raises IOError: The path of the pickled figure. 
    :return: a matplotlib figure 
    :rtype: plt.Figure
    """
    try:
        with open(file_path, 'rb') as reader_buf:
            fig=pickle.load(reader_buf)
        return fig 
    except Exception as exp: 
        raise IOError(f'While loading your figure, the following error was encountered: {exp}')   

def build_sequence_table(sequence_dict:Dict[str,str])->pd.DataFrame:
    """construct a sequences database from a sequences dict object  
    
    :param sequence_dict: a dict that contain the protein ids as keys and sequences as values. 
    :type sequence_dict: Dict[str,str]
    :return: pandas dataframe that contain the protein ID and the associated protein sequence 
    :rtype: pd.DataFrame
    """
    return pd.DataFrame(sequence_dict)

def get_idx_peptide_in_sequence_table(sequence_table:pd.DataFrame, peptide:str)->List[str]:
    """check the sequences table if the provided peptide is locate in one of its sequences and returns 
    a list of protein identifiers containing the identifier of the hit proteins.

    :param sequence_table:  pandas dataframe that contain the protein ID and the associated protein sequence 
    :type sequence_table: pd.DataFrame
    :param peptide: The peptide sequence to query the protein with 
    :type peptide: str
    :return: A list of protein identifiers containing the identifier of the hit proteins
    :rtype: List[str]
    """
    if sequence_table.columns != ['Sequences']: 
        sequence_table.columns=['Sequences']
    return sequence_table.loc[sequence_table['Sequences'].str.contains(peptide)].index.tolist()

def check_peptide_made_of_std_20_aa(peptide:str)->str:
    """Check if the peptide is made of the standard 20 amino acids, if this is the case, 
    it return the peptide sequence, otherwise it return an empty string
    
    :param peptide: a peptide sequence to check its composition
    :type peptide: str
    :return: True, if the peptide is made of the standard 20 amino acids, False otherwise. 
    :rtype: str
    """
    amino_acids=['A','R','N','D','C','E','Q','G','H','I','L','K','M','F','P','S','T','W','Y','V']
    for amino_acid in peptide: 
        if amino_acid not in amino_acids: 
            return ''
    return peptide

def get_experiment_summary(ident_table: pd.DataFrame) -> pd.DataFrame:
    """takes as an input an identification table and return a summary table containing the count of unique peptides,\
    unique proteins, maximum peptide length, minmum peptide length, median and mean peptide length 

    :param ident_table: the identification table as returned by one of the parser functions defined in the IO modules 
    :type ident_table: pd.DataFrame
    :return: The summary table 
    :rtype: pd.DataFrame
    """
    # compute the statsitics 
    num_unique_peptides: int = len(set(ident_table.peptide))
    num_unique_proteins: int = len(set(ident_table.protein))
    peptide_len: np.ndarray= np.array(ident_table.end_index) - np.array(ident_table.start_index)
    max_len: int = np.max(peptide_len) 
    min_len: int = np.min(peptide_len)
    medain_len: float = np.median(peptide_len)
    mean_len: float = np.mean(peptide_len)
    # generate the data frame 
    res: pd.DataFrame = pd.DataFrame({
        'num_unique_peptides':[num_unique_peptides], 
        'num_unique_proteins': [num_unique_proteins], 
        'max_len': [max_len], 
        'min_len':[min_len],
        'median_len':[medain_len], 
        'mean_len': [mean_len]
    })
    # return the results 
    return res 

def combine_summary(child_dfs: List[pd.DataFrame], root_df: pd.DataFrame = None) -> pd.DataFrame: 
    """combine multiple summaray dataframes into one dataframe 

    :param child_dfs: a list of summary dataframes to conctinate into one 
    :type child_dfs: List[pd.DataFrame]
    :param root_df: a dataframe to append the child dataframe to its tail, defaults to None
    :type root_df: pd.DataFrame, optional
    :return: a dataframe containing the root and the child dataframes 
    :rtype: pd.DataFrame
    """
    if root_df is None: 
        root_df= pd.DataFrame(columns=['num_unique_peptides','num_unique_proteins','max_len','min_len','median_len','mean_len'])
    # loop over the child dataframes
    for child_df in child_dfs: 
        root_df=pd.concat([root_df, child_df], axis=0)
    # return the results 
    return root_df


#!/usr/bin/env python
"""A representation of the Tissue used in an IP Experiment. 
"""
# load the modules 
from __future__ import annotations
import pandas as pd
import numpy as np 
from typing import Union, List, Dict 
from IPTK.Classes.Database import CellularLocationDB, GeneExpressionDB
# define the tissue class 
class ExpressionProfile: 
	"""a representation of tissue reference expression value. 
	"""
	def __init__(self, name: str,
	 	expression_table: pd.DataFrame, 
		aux_proteins: pd.DataFrame = None)->ExpressionProfile: 
		"""Create an expression profile instance from an expression table. 

		:param name: the name of the tissue 
		:type name: str
		:param expression_table: A dataframe with the following three columns gene id, gene name, and the expression value.   
		:type expression_table: pd.DataFrame
		:param aux_proteins: A table that contain the expression table of auxillary proteins that does not belong to the tissue per sa,
		for example, pathogen-derived genes, defaults to None
		:type: pd.DataFrame, optional
		:raises ValueError: if the tissue name is None
		:raises ValueError: if the expression table is not a pandas dataframe instance 
		:raises ValueError: if the expression table does not have the correct shape 
		:raises ValueError: if the aux_proteins does not have the correct shape
		:return: an expression profile instance 
		:rtype: ExpressionProfile
		"""
		if name is not None:
			self._name=name
		else:
			raise ValueError(f'The tissue name can not be None')
		
		if not isinstance(expression_table, pd.DataFrame):
			raise ValueError(f'Expression table must be a pd.DataFrame instance, however, your input is of type: {type(expression_table)}')
		
		if expression_table.shape[1] != 3:  
			raise ValueError(f"The provided expression table must have three columns, however, your table have: {expression_table.shape[1]}")
		
		self._exp_map=expression_table
		# set the columns name 
		self._exp_map.columns=['gene', 'gene_name','exp_value']
		# parse the auxillary table 
		if aux_proteins is not None:
			if aux_proteins.shape[1]!=3:
				raise ValueError(f'The provided auxillary proteins table must have three columns, however, your table have: {aux_proteins.shape[1]}')
			aux_proteins.columns=['gene', 'gene_name','exp_value']
			# concatenate the results 
			self._exp_map=pd.concat([self._exp_map,aux_proteins],axis=0)
		return

	def get_gene_id_expression(self,gene_id: str)-> float: 
		"""
		:param gene_id: the gene id to retrive its expression value from the database
		:type gene_id: str
		:raises KeyError: if the provided id is not defined in the instance table 
		:return: the expression value of the provided gene id.
		:rtype: float
		"""
		if gene_id not in self._exp_map.iloc[:,0].tolist(): 
			raise KeyError(f"The provided gene id: {gene_id} is not defined in the database")
		return self._exp_map.loc[self._exp_map.iloc[:,0]==gene_id,:].iloc[0,2]
	
	def get_gene_name_expression(self,gene_name: str)-> float: 
		"""
		:param gene_name: the gene name to retrive its expression value from the database
		:type gene_name: str
		:raises KeyError: if the provided id is not defined in the instance table 
		:return: the expression value of the provided gene name. 
		:rtype: float
		"""
		if gene_name not in self._exp_map.iloc[:,1].tolist(): 
			raise KeyError(f"The provided gene name: {gene_name} is not defined in the database.")
		return self._exp_map.loc[self._exp_map.iloc[:,1]==gene_name,:].iloc[0,2]

	def get_name(self)->str:
		"""
		:return: the name of the tissue where the expression profile was obtained
		:rtype: str
		"""
		return self._name

	def get_table(self)->pd.DataFrame:
		"""
		:return: return a table that contain the expression of all the transcripts in the current profile \
		including core and auxiliary proteins
		:rtype: pd.DataFrame
		"""
		return self._exp_map
	
	def __len__(self)->int: 
		"""
		:return: return the number of unique transcripts in the profile 
		:rtype: int
		"""
		return len(set(self._exp_map.shape[0]))

	def __str__(self)->str:
		""" 
		:return: a string representation for the class instance  
		:rtype: str
		"""
		return f'{self._name} with an expression profile covering {self._exp_map.shape[0]} genes.'
	
	def __repr__(self)->str:
		return str(self)

class Tissue:
	def __init__(self, name: str, main_exp_value: GeneExpressionDB, 
	main_location: CellularLocationDB, aux_exp_value: GeneExpressionDB = None,
	aux_location: CellularLocationDB = None) -> Tissue:
		"""The initializer of the Tissue class 

		:param name: The name of the tissue 
		:type name: str
		:param main_exp_value: A GeneExpressionDB instace containing the gene expression accross different tissues 
		:type main_exp_value: GeneExpressionDB
		:param main_location: A CellularLocationDB instance that contain the sub cellular locations for the proteins expressed in the tissue
		:type main_location: A CellularLocationDB 
		:param aux_exp_value: A GeneExpressionDB instance that contain the expression table of auxillary proteins that does not belong to the tissue per sa,
	  	for example, pathogen-derived genes or extra-cellular matrix, defaults to None.
		:type aux_exp_value: GeneExpressionDB, optional
		:param aux_location: CellularLocationDB instance that contain the sub cellular locations for proteins that does not belong to the tissue of interest per sa
	  	for example, pathogen-derived proteins or media-added proteins, defaults to None
		:type aux_location: CellularLocationDB, optional
		:raises KeyError: if the provided tissue name is not defined in the gene expression database 
		:return: [description]
		:rtype: Tissue
		"""
		if name not in main_exp_value.get_tissues():
			raise KeyError(f'The provided tissue name: {name} is not in the provided main expression database')
		# add the expression profile 
		if aux_exp_value is not None: 
			self._exp_prof: ExpressionProfile = ExpressionProfile(name=name,
			expression_table=main_exp_value.get_expression_in_tissue(name),
			aux_proteins=aux_exp_value.get_table())
		else: 
			self._exp_prof: ExpressionProfile = ExpressionProfile(
				name=name, expression_table=main_exp_value.get_expression_in_tissue(name))	
		# add the cellular location profile:  
		self._cell_loc = main_location
		if aux_location is not None:
			self._cell_loc.add_to_database(aux_location)
	
	def get_expression_profile(self)->ExpressionProfile: 
		"""
		:return: the expresion profile of the current tissue 
		:rtype: ExpressionProfile
		"""
		return self._exp_prof
	
	def get_name(self)->str:
		"""
		:return: the name of the tissue 
		:rtype: str
		"""
		return self._exp_prof.get_name()

	def get_subCellular_locations(self) ->CellularLocationDB:
		""" 
		:return: the sub-cellular localization of all the proteins stored in current instance resources. 
		:rtype: CellularLocationDB
		"""
		return self._cell_loc
	
	def __str__(self)->str:
		"""
		:return: a string representation for the current instance
		:rtype: str
		"""
		return f'{self._exp_prof.get_name()} with an associated expression profile covering: {len(self._exp_prof)} genes and a sub-cellular location covering: {len(self._cell_loc)} genes '
		


#!/usr/bin/env python 
"""A representation of the eluted peptides and its identified proteins.
"""
# Load the modules
from __future__ import annotations
import numpy as np 
import pandas as pd
from typing import List 
from IPTK.Utils.Types import Range, Sequences, MappedProtein, Organisms
# define the class 
class Peptide: 
	"""An representation of an eluted peptide. 
	"""
	def __init__(self,pep_seq: str)->Peptide:
		"""class constructor 
		:param pep_seq: the peptide sequence 
		:type pep_seq: str
		:return: a Peptide instance 
		:rtype: Peptide
		"""
		self._peptide=pep_seq
		self._parent_proteins=dict()

	def get_length(self) -> int:
		"""
		:return: the length of the peptide
		:rtype: int
		"""
		return len(self._peptide) 

	def get_peptide_seq(self) -> str: 
		"""
		:return: the sequence of the peptide.
		:rtype: str
		"""
		return self._peptide

	def add_parent_protein(self, parent_protein, 
						start_index: int, end_index: int) -> None:
		"""add a protein instance as a parent to the current peptide. 
		The library use Python-based indexing where its 0-indexed and ranges are treated as [start, end). 
		:param parent_protein: a Protein instance that act as a parent to the peptide.
		:type parent_protein: Protein 
		:param start_index: the position in the parent protein where the peptide starts
		:type start_index: int 
		:param end_index: the index of the amino acid that occurs after the last amino acid in the peptide, 
		:type start_index: int 
		"""
		self._parent_proteins[parent_protein.get_id()]={
		'protein':parent_protein, 
		'start_index':start_index, 
		'end_index':end_index}

	def get_number_parent_protein(self) -> int:
		"""
		:return: The number of parent proteins this instance has 
		:rtype: int
		"""
		return len(self._parent_proteins) 

	def get_flanked_peptide(self,flank_len:int)-> Sequences: 
		"""
		:param flank_len: The length of the flanking regions 
		:type flank_len: int
		:return: A list of string containing the length of the peptide + the flanking region from \
		both the N and C terminal of the instance peptide, from all proteins.
		:rtype: Sequences
		"""
		# if the parent has no parents, the function returns an empty list. 
		if self.get_number_parent_protein() == 0: 
			return [''] 
		# else: 
		# iterate over all the elements to obtain the sequences 
		results=[]
		for prot_id in self._parent_proteins.keys():
			parent_protein=self._parent_proteins[prot_id]
			protein_seq=parent_protein['protein']
			# check that the flaking region + protein length is within the protein length.
			if parent_protein['start_index'] - flank_len < 0: 
				start_idx=0
			else: 
				start_idx=parent_protein['start_index'] - flank_len
			# check the C-terminal side 
			if flank_len+parent_protein['end_index'] > len(protein_seq): 
				end_idx=len(protein_seq) 
			else: 
				end_idx= flank_len+parent_protein['end_index']	
			results.append(protein_seq[start_idx:end_idx])
		return results

	def map_to_parent_protein(self) ->MappedProtein:
		""" Mapped the instance peptide to the parent protein and returned a 
		list of Numpy arrays where each array has a size of 1 by protein length. 
		within the protein the range representing the peptide is encoded as one while
		the rest is zero. 

		:return: A list of binary encoded arrays represent this mapping. 
		:rtype: MappedProtein
		"""
		# if the peptide has no associated parent protein yet, 
		# return an empty list 
		if self.get_number_parent_protein() == 0: 
			return [np.zeros(shape=(0,0))] 
		# else we allocate a list to compute the results 
		results=[]
		# compute the results for each parent protein
		for prot in self._parent_proteins.keys(): 
			pp= self._parent_proteins[prot]
			temp_array=np.zeros(shape=(1,len(pp['protein'])))
			temp_array[0,pp['start_index']:pp['end_index']]= 1
			results.append(temp_array) 
		return results 

	def get_non_presented_peptides(self, length: int )->Sequences: 
		"""
		:param length: The length of the non-presented peptides
		:type length: int
		:return: non-presented peptides from all the parent protein of the current peptide instance.  
		:rtype: Sequences
		"""
		results=[]
		for pro in self._parent_proteins.keys():
			results.append(self._parent_proteins[pro]['protein'].get_non_presented_peptide(self._parent_proteins[pro]['start_index'],self._parent_proteins[pro]['end_index'], length=length))
		return results
	
	def get_parent_proteins(self)->List[str]:
 		return list(set(self._parent_proteins.keys())) 

 	
	def is_child_of(self,pro_id:str)->bool:
		"""
		:param pro_id: is the protein id 
		:type pro_id: str
		:return: True if a user provided protein-id is a parent for the instance peptide, False otherwise 
		:rtype: bool
		"""
		return pro_id in self.get_parent_proteins()
	
	def get_pos_in_parent(self,pro_id:str)->Range:
		"""
		:param pro_id: the id of the parent protein 
		:type pro_id: str
		:raises ValueError: If the identifier is not a parent of the instance 
		:return: the start and end position of the instance peptide in the parent with the provided identifier
		:rtype: Range
		"""
		if not self.is_child_of(pro_id):
			raise ValueError(f"The provided protein id: {pro_id} is not a parent of the current peptide")
		# get the start and end elements 
		start=self._parent_proteins[pro_id]['start_index']
		end=self._parent_proteins[pro_id]['end_index']
		# return the results 
		return start, end 
	
	def get_parent(self,pro_id:str):
		"""
		:param pro_id: The protein identifer 
		:type pro_id: str
		:return: The parent protein that has an id matching the user defined pro_id
		:rtype: Protein
		"""
		if self.is_child_of(pro_id):
			return self._parent_proteins[pro_id]['protein']

	def get_number_of_parents(self)->int:
		"""
		:return: The number of instance parent proteins
		:rtype: int
		"""
		return len(self._parent_proteins)
	
	def get_n_terminal_flank_seq(self, flank_len : int)->List[str]:
		"""
		:param flank_len: the length of the flanking regions 
		:type flank_len: int
		:return: A list of string containing the sequences located upstream of the peptide in the parent protein. 
		:rtype: List[str]
		"""
		# if the peptide has no parent -> return a list with an empty string 
		if self.get_number_of_parents()==0:
			return ['']	
		#  else, we declare a list to allocate the results to it ==>  
		results=[]
		# loop over all the parent to obtain the upstream sequences 
		for parent in self._parent_proteins.keys(): 
			parent_protein=self._parent_proteins[parent]
			start_index=parent_protein['start_index']
			if start_index-flank_len<0: 
				results.append(parent_protein['protein'][0:start_index])
			else: 
				results.append(parent_protein['protein'][start_index-flank_len:start_index])
		# return the results 
		return results 
		 
	def get_c_terminal_flank_seq(self, flank_len: int)->List[str]:
		"""
		:param flank_len: The length of the flanking regions 
		:type flank_len: int
		:return: A list of string containing the sequences located downstream of the peptide in the parent protein. 
		:rtype: [type]
		"""
		# if the peptide has no parent -> return a list with an empty string 
		if self.get_number_of_parents()==0:
			return ['']	
		#  else, we declare a list to allocate the results to it ==>  
		results: List[str]=[]
		# loop over all the parent to obtain the upstream sequences 
		for parent in self._parent_proteins.keys(): 
			# get the parent protein 
			parent_protein=self._parent_proteins[parent]
			# get the end of the protein 
			end_index=parent_protein['end_index']
			# get the protein length  
			protein_length=len(parent_protein['protein']) 		
			if end_index+flank_len>protein_length: # ==> avoid indexing outside of the parent protein
				results.append(parent_protein['protein'][end_index:protein_length])
			else: 
				results.append(parent_protein['protein'][end_index:end_index+flank_len])
		# return the results 
		return results 

	def add_org_2_parent(self,prot_name: str,org:str)->None:
		"""adds the source organism of one of the instance parent protein

		:param prot_name: The name of the protein, i.e. the identifier of the protein 
		:type prot_name: str
		:param org: The name of the organism 
		:type org: str
		:raises ValueError: Incase the provided protein is not a parent of the provided peptide 
		"""
		if not self.is_child_of(prot_name):
			raise ValueError(f'The provided protein id is not a parent of the current peptide')
		self._parent_proteins[prot_name]['protein'].set_org(org)
	
	def get_parents_org(self)->Organisms:
		"""
		:return: A list containing the name of each parent protein source organisms
		:rtype: Organisms
		"""
		results=[]
		for protein in self._parent_proteins.keys():
			results.append(self._parent_proteins[protein]['protein'].get_org())
		return results 
	
	def __len__(self)->int:
		"""
		:return: The length of the peptide 
		:rtype: int
		"""
		return len(self._peptide)
	
	def __getitem__(self,aa_index:int)->str:
		"""Bracket based indexing into the peptide sequence

		:param aa_index: the amino acid index in the peptide sequence, for example 3rd amino acid
		:type aa_index: int
		:return: the amino acid correspond to the provided index  
		:rtype: str
		"""
		return self._peptide[aa_index]
	
	def __str__(self)->str:
		"""
		:return: The sequence of the peptide 
		:rtype: str
		"""
		return self._peptide

	def __repr__(self)->str:
		"""
		:return: A string representation for the class 
		:rtype: str
		"""
		return f'A Peptide instance with {self.get_number_of_parents()} parent protein'
	
		
		
		
		


#!/usr/bin/env python 
"""A description for an IP proband 
"""
# Load the modules 
from __future__ import annotations
from IPTK.Utils.UtilityFunction import generate_random_name
# generate the class 
class Proband: 
    def __init__(self,**info)->Proband:
        """create A Proband using an arbitrary number of key-value pairs containing information about the proband
        :return: A proband instane 
        :rtype: Proband
        """
        self._meta_info=dict()
        for key, value in info.items(): 
            self._meta_info[key]=value 
        keys=[temp_key.lower() for temp_key,_ in info.items()]
        if 'name' not in keys: 
            self._meta_info['name']=generate_random_name(12)
    # add meta info 
    def update_info(self, **info)->None:
        """Add new or update existing info about the patient using an arbitrary number of key-value pairs to be added to the instance meta-info dict
        """
        for key, value in info.items(): 
            self._meta_info[key]=value
            
    def get_name(self)->str:
        """
        :return: The name of the proband
        :rtype: str
        """
        return self._meta_info['name']
    
    def get_meta_data(self)->dict:
        """
        :return: A dict containing all the meta-data about the proband 
        :rtype: dict
        """
        return self._meta_info


#!/usr/bin/env python 
r"""The class provides methods for visualizing different aspects of the protein \
    biology. This is achieved through three main methods: 
        1- add_segmented_track: which visualize information about \
        non-overlapping protein substructures, for example, protein domains.

        2- add_stacked_track: which visualize information about \
        overlapping protein substructures, for example, splice variants. 
        
        3- add_marked_positions_track: which visualize or highlight \
        positions in the protein, for example, sequence variants, or PTM. 
    
    The class also provides functions for visualizing the relationship between 
    a protein and its eluted peptide/peptides in an analogous manner to the way 
    NGS reads are aligned to genomic regions. This can be useful to identify  
    regions in the protein with high/low number of eluted peptides, i.e.,Coverage.
    Also, to link it with other facests of the protein like domain organization,\
    PTM, sequence/splice variants. 
    
    Notes
    -----
    each figure should have a base track this can be done explicitly by calling
    the function add_base_track or by implicitly by calling the function 
    add_coverage_plot with the parameter coverage_as_base=True.

"""
# load the modules 
from __future__ import annotations
import numpy as np 
import matplotlib.pyplot as plt 
import os 
from typing import Tuple, List, Dict, Union
# define the class 
class Annotator:
    r""" A high level API to plot information about the protein, for example, PTM, Splice variant etc, using matplotlib library 
    """
    def __init__(self,protein_length: int,figure_size:Tuple[int,int],figure_dpi:int,
                 face_color="white")->Annotator:
        """
        Parameters
        ----------
        protein_length : int 
            the length of the protein.
        figure_size : tuple/list 
            The figure size in inches. The first element in the tuple/list is 
            the width and the second is the height.
        figure_dpi : int
            the dots per inch, which control the figure resolution.
        face_color : string, optional
            the bachground colors . The default is "white".
        Returns
        -------
        Annotator.

        """
        #check the user input
        if protein_length<0:
            raise ValueError("Protein length must be bigger than zero")

        if len(figure_size)!=2: 
            raise ValueError("figure_size must be a tuple or list with two elements")
        # create the figure 
        self.fig=plt.figure(figsize=figure_size,dpi=figure_dpi,
                            facecolor=face_color) 
        # add the protein length to self
        self.protein_length=protein_length
        self.figure_dpi=figure_dpi
        self.figure_size=figure_size
        return
    
    def add_base_track(self, space_fraction:float=0.3,
                       protein_name_position:float=0.5,
                       track_label:str="base_track",
                       track_label_dict: Dict[str,Union[int,str]]={"fontsize":8,"color":"black"},
                       protein_name: str="A protein", 
                       protein_name_dict:Dict[str,Union[int,str]] ={"fontsize":10,"color":"black"},
                       rect_dict: Dict[str,Union[int,str]]={"color":"olive","capstyle":"butt"},
                       number_ticks: int =10,
                       xticks_font_size: int =4,
                          ):
        """
        Description
        -----------
        Adds a base track to the figure. 

        Parameters
        ----------
        space_fraction : float, optional
            A float between 0 and 1 that represent the fraction of space left 
            below and above the track. The default is 0.3 which means that the
            track will be drown on a 40% while 60% are left as an empty space
            below and above the track.
            
        protein_name_position : float, optional
            A float between 0 and 1 which control the relative position of the 
            protein name on the y-axis. The default is 0.5.
            
        track_label : string, optional
            The name on the track, which will be shown on the y-axis. 
            The default is "base_track".
            
        track_label_dict : Dict[str,Union[int,str]], optional
            The parameters that control the printing of the track_label,
            for example, the font size and the color. These parameters should 
            be provided as dict that will be fed to the function 
            ``axes.set_ylabel``.The default is {"fontsize":8,"color":"black"}.
            
        protein_name : string, optional
            The name of the protein to be printed to the track.
            The default is "A protein".
            
        protein_name_dict : Dict[str,Union[int,str]], optional
            the parameters that control the printing of the protein name,
            for example, the font size and the color. These parameters should 
            be provided as dict that will be fed to the function 
            ``axes.text()``. The default is {"fontsize":10,"color":"black"}.
            
        rect_dict : Dict[str,Union[int,str]], optional
            a dictionary that control the character of the track itself, for example,
            the color and the transparency. this dict will be fed to the function
            ``plt.Rectangle()``. 
            The default is {"color":"olive","capstyle":"butt"}.
        
        number_ticks:int
            The number of ticks on the x-axis. The default is 10.
        
        xticks_font_size: int
            The font size of the x-axis ticks. The default is 4.
            
        Returns
        -------
        None.
        
        Examples
        --------
        >>> example_1=VisTool(250,(3,5),300) 
            # create a graph of size 3 inches by 5 inches with a 300 dots per
            # inch (DPI) as a resolution metric for a protein of length 250 amino acids
        
        >>> example_1.add_base_track() 
            # adds a basic track using the default parameters. 
        
        >>> example_1.add_base_track(space_fraction=0.1,
                                    track_label="example_1",
                                    track_label_dict={"fontsize":5,"color":"blue"}
                                    number_ticks=5,
                                    xticks_font_size=6)
            # generate a base track with 10% empty space above and below 
            #  the track. Track will have the name example_1 and it will be 
            # shown in font 5 instead of 8 and in blue color instead of black.
            # five ticks will be shown on the x-axis using a font of size 6.
        
        Notes
        -----
            calling the function more than once will result in an overriding of
            the previously added base track, for example, in the examples 
            section calling add_base_track for the second time will overrides 
            the graph build by the previous call. 

        """
        # add a subplot to the figure
        ax=self.fig.add_subplot(111)
        # set the axis limits for the axes
        ax.axis(xmin=0,xmax=self.protein_length,ymin=0,
                ymax=1)
        # build the protein representation
        rectangle=plt.Rectangle((0,space_fraction),
            width=self.protein_length,height=1-(2*space_fraction),**rect_dict)
        # add it to the axes
        ax.add_patch(rectangle)
        # add the protein name 
        ax.text(x=self.protein_length/3,
                y=protein_name_position,s=protein_name,**protein_name_dict)
        # remove the spines of the axes
        ax.spines["left"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.spines["bottom"].set_visible(False)
        # flip the x axes
        ax.xaxis.tick_top()
        # adjust the ticks of the x-axis
        step_size=int(np.floor(self.protein_length/number_ticks))
        ax.set_xticks(ticks=np.arange(0,self.protein_length,step_size),minor=False)
        ax.tick_params(axis='x', which='major', labelsize=xticks_font_size)
        # remove the y-label ticks
        ax.get_yaxis().set_ticks([])
        # set the y-label
        ax.set_ylabel(track_label,**track_label_dict)
    
    def add_segmented_track(self,track_dict:Dict[str,Dict[str,Union[int,str]]],
                            track_label: str ="A segmented Track",
                            track_label_dict: Dict[str,Union[int,str]]={"fontsize":8,"color":"black"},
                            track_element_names_dict:Dict[str,Union[int,str]]={"fontsize":8,"color":"black"},
                            center_line_dict:Dict[str,Union[int,str,float]]={"alpha":0.5,"linewidth":0.5},
                            track_elements_dict: Dict[str,Union[int,str]]={"color":"brown","capstyle":"butt"},
                            show_names: bool=True)->None:
        """Adds a segmentation track which show non-overlapping features of the protein. 
        
        Parameters
        ----------
        track_dict : Dict[str,Dict[str,Union[int,str]]]
            A dict that contain the non-overlapping features of the protein. 
            The dict is assumed to have the following structure: a dict with the
            feature_index as a key and associated features as values. The associated
            features is a dict with the following three keys:
                
                1- Name: which contain the feature name

                2- startIdx: which contain the start position of the protein
                
                3- endIdx: which contain the end position of the protein
                
        track_label : str, optional
            The name of the track, which will be shown on the y-axis. 
            The default is "A segmented Track". 
            
        track_label_dict : Dict[str,Union[int,str]], optional
            the parameters that control the printing of the track_label,
            for example, the font size and the color. These parameters should 
            be provided as dict that will be fed to the function 
            ``axes.set_ylabel``. The default is {"fontsize":8,"color":"black"}.
            
        track_element_names_dict : Dict[str,Union[int,str]], optional
            the parameters that control the printing of the feature names on the track,
            for example, the font size and the color. These parameters should 
            be provided as a dict that will be fed to the function ``axes.text``. 
            The default is {"fontsize":8,"color":"black"}.

        center_line_dict: Dict[str,Union[int,str, float]], optional 
            The parameters that control the printing of the center line of a segmented track object.
            The default is {"fontsize":8,"color":"black"}.
            
        track_elements_dict : Dict[str,Union[int,str]], optional
            the parameters that control the printing of the feature 
            rectangluar representation for example the color, the dict will be fed 
            to the function ``plt.Rectangle``. 
            The default is {"color":"brown","capstyle":"butt"}.
        
        show_names : bool, optional
            whether or not to show the name of the features. The default is True.

        Returns
        -------
        None.
        
        Examples
        --------
        >>> test_dict={"domain1":{"Name":"domain_one","startIdx":55,"endIdx":150},
                       "domain2":{"Name":"domain_Two","startIdx":190,"endIdx":225}}
        # first define a dict object that define some protein features.

        
        >>> example_1=Annotator(protein_length=250, figure_size=(5,3), figure_dpi=200)
        # creating a Annotator instance
        
        >>> example_1.add_base_track()
        # add a base_track

        
        >>> example_1.add_segmented_track(test_dict) # build a segmented track using the default parameters
        # add the segmented track

       
        >>> example_1.add_segmented_track(track_dict=test_dict,
                                          track_label="Domains",
                                          track_elements_dict={"color":"brown"})
        # add a second segmented track with track name set to Domains and elements 
        # of the track shown as brown rectangles.
        
        Notes
        -----
        Any panel can have one or more segmented-tracks. Thus, in the above examples
        calling the method ``add_segmented_track`` for the second time does NOT
        override the previous segmented track it create a new one and added to the
        figure.
        
        """
        # change the geometry of the graph to adapt to the added axes
        number_axes=len(self.fig.axes)
        for i in range(number_axes):
            self.fig.axes[i].change_geometry(number_axes+1,1,i+1)
        ax=self.fig.add_subplot(number_axes+1,1,number_axes+1)
        # set the axes axis limits
        height=1
        ax.axis(xmin=0,xmax=self.protein_length,ymin=0,ymax=height)
        ax.axhline(y=0.45*height,xmin=0,xmax=self.protein_length,color="black", **center_line_dict)
        # get the position of the dictionary objects
        for key in track_dict.keys():
            segment_length=track_dict[key]["endIdx"]-track_dict[key]["startIdx"]
            ax.add_patch(plt.Rectangle((track_dict[key]["startIdx"],0.33*height), 
                width=segment_length, height=0.33*height,**track_elements_dict))
            if show_names:
                ax.text(
                    x=track_dict[key]["startIdx"],
                    y=0.5*height,s=track_dict[key]["Name"],
                    **track_element_names_dict)
        
        # adjust the figure of the track
        ax.spines["left"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.spines["bottom"].set_visible(False)
        # remove the ticks from the x and y axis
        ax.get_xaxis().set_ticks([])
        ax.get_yaxis().set_ticks([])
        # add the track name to the libray
        ax.set_ylabel(track_label,**track_label_dict)
        # set the position of the trach

    def add_stacked_track(self,track_dict: Dict[str,Dict[str,Union[int,str]]],
                            track_label: str ="A stacked Track",
                            track_label_dict: Dict[str,Union[int,str]] ={"fontsize":8,"color":"black"},
                            track_element_names_dict: Dict[str,Union[int,str]]={"fontsize":8,"color":"black"},
                            track_elements_dict: Dict[str,Union[int,str]]={"color":"magenta","capstyle":"butt"},
                            base_line_dict: Dict[str,Union[int,str]]={"color":"black","linewidth":1},
                            show_names:bool =True):
        """
        Description
        -----------
        The function adds a stacked_track to a visualization panel. 
        The stacked track is used to show overlapping protein features, for example,
        different splice variants. 

        Parameters
        ----------
        track_dict : Dict[str,Dict[str,Union[int,str]]]
            A dict that contain the overlapping features of the protein. 
            The dict is assumed to have the following structure, a dict with the
            feature_index as a key and associated features as values. The associated
            features is a dict with the following three keys:

                1- Name: which contain the feature's name
                
                2- startIdx: which contain the start position of the feature.
                
                3- endIdx: which contain the end position of the feature.
                
       
        track_label : str, optional
            The name of the track, which will be shown on the y-axis. 
            The default is "A stacked Track".
            
        track_label_dict : Dict[str,Union[int,str]], optional
            the parameters that control the printing of the track_label,
            for example, the font size and the color. These parameters should 
            be provided as dict that will be fed to the function 
            ``axes.set_ylabel``.The default is {"fontsize":8,"color":"black"}.
            
        track_element_names_dict : Dict[str,Union[int,str]], optional
            the parameters that control the printing of the feature names on the track,
            for example, the font size and the color. These parameters should 
            be provided as a dict that will be fed to the function ``axes.text``. 
            The default is {"fontsize":8,"color":"black"}.
            
        track_elements_dict : Dict[str,Union[int,str]], optional
            the parameters that control the printing of the feature 
            rectangluar representation for example the color, the dict will be fed 
            to the function ``plt.Rectangle``.
            The default is {"color":"magenta","capstyle":"butt"}.
            
        base_line_dict : Dict[str,Union[int,str]], optional
            the parameters that control the shape of the base line,
            for example, color and/or line width. These parameters are going to be fed
            to the function ``axes.hlines``. 
            The default is {"color":"black","linewidth":1}.
            
        show_names : bool, optional
            whether or not to show the name of the features. The default is True.

        Returns
        -------
        None.
        
        Examples
        --------
        >>> test_dict={"feature_1":{"Name":"X","startIdx":55,"endIdx":150},
                       "feature_2":{"Name":"Y","startIdx":85,"endIdx":225},
                       "feature_3":{"Name":"Z","startIdx":160,"endIdx":240}}
         # first define a dict object that define some protein features.
        
        
        >>> example_1=Annotator(protein_length=250, figure_size=(5,3), figure_dpi=200)
        # creating a Annotator instance

        
        >>> example_1.add_base_track()
        # add a base_track

        
        >>> example_1.add_segmented_track(test_dict) # build a stacked track using the default parameters.
        # add the stacked track
       
        >>> example_1.add_segmented_track(track_dict=test_dict,
                                          track_label="OverLappingFeat",
                                          track_elements_dict={"color":"red"})
        # add a second segmented track with track name set to OverLappingFeat and elements 
        # of the track shown as red rectangles.
        
        Notes
        -----
        Any panel can have zero, one or more than one stacked-track.
        Thus, in the above examples calling the method ``add_stacked_track`` 
        for the second time does NOT override the previous stacked track 
        it creates a new one and added to the figure.
        """
        # change the geometry to accommodate a new figure
        number_axes=len(self.fig.axes)
        for i in range(number_axes):
            self.fig.axes[i].change_geometry(number_axes+1,1,i+1)
        # set the base height
        height=1
        # create a new axes:
        ax=self.fig.add_subplot(number_axes+1,1,number_axes+1)
        # adjust the x and y axis limits
        ax.axis(xmin=0,xmax=self.protein_length,ymin=0,ymax=height)
        # add a base line
        ax.hlines(y=0,xmin=0,xmax=self.protein_length,**base_line_dict)
        # adjust the space between the stacked and the formating characters
        number_of_stacked_tracked=len(track_dict)
        hight_per_track=height/number_of_stacked_tracked
        space_margine=0.2*hight_per_track # 0.2 below and above the graph will 
        track_hight=0.6*hight_per_track
        text_hight=0.5*hight_per_track
        track_elements=list(track_dict.keys())
        # plotting the tracks
        for idx in range(len(track_elements)):
            start_index=track_dict[track_elements[idx]]["startIdx"]
            end_index=track_dict[track_elements[idx]]["endIdx"]
            width=end_index-start_index
            name=track_dict[track_elements[idx]]["Name"]
            ax.add_patch(plt.Rectangle(
                (start_index,(idx*hight_per_track)+space_margine),
                width=width,height=track_hight,**track_elements_dict
                ))
            if show_names:
                ax.text(x=start_index,y=(idx*hight_per_track)+text_hight,
                    s=name,**track_element_names_dict)
        # processes the axes:
        ax.spines["left"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.spines["bottom"].set_visible(False)
        ax.get_yaxis().set_ticks([])
        ax.get_xaxis().set_ticks([])
        ax.set_ylabel(track_label,**track_label_dict)
    
    def add_marked_positions_track(self,positions:List[int],
                    height_frac: float =0.5,
                    marker_bar_dict: Dict[str,Union[int,str]] ={"color":"black","linestyles":"solid"},
                    marker_dict: Dict[str,Union[int,str]] ={"color":"red","s":3},
                    track_label: str ="A marked positions Track",
                    track_label_dict: Dict[str,Union[int,str]] ={"color":"black","fontsize":8},
                    base_line_dict: Dict[str,Union[int,str]] ={"color":"black","linewidth":1}):
        """
        Description
        -----------
        The function adds a marked position to the track which is shown to highlight
        certain amino acid position within the protein, for example, a sequence
        variant position, or PTM position.

        Parameters
        ----------
        positions : List[int]
            a list that contain the position/positions that should be heighlighted
            in the protein sequence.
        
        height_frac : float
            the relative hight of the marked positions. The default is 0.5 which
            means that the hight of the marker will be 50% of the y-axis height.
            
        marker_bar_dict : Dict[str,Union[int,str]], optional
            The parameters of the marker position bar, for example, line width
            or color. These parameters are going to be fed to the function
            ``plt.hlines``. The default is {"color":"black","linestyles":"solid"}.
            
        marker_dict : Dict[str,Union[int,str]], optional
            These are the parameters for the marker points which sits on top of the marker bar,
            for example, the color, the shape or the size. 
            The default is {"color":"red","s":3}.
            
        track_label : str, optional
             The name of the track, which will be shown on the y-axis.
             The default is "A marked positions Track".
            
        track_label_dict : Dict[str,Union[int,str]], optional
            The parameters that control the printing of the track_label,
            for example, the font size and the color. These parameters should 
            be provided as dict that will be fed to the function 
            ``axes.set_ylabel``.The default is {"fontsize":8,"color":"black"}.
            
       base_line_dict : Dict[str,Union[int,str]], optional
            The parameters that control the shape of the base line, for example, color and/or line width. These parameters are going to be fed
            to the function ``axes.hlines``. 
            The default is {"color":"black","linewidth":1}.

        Returns
        -------
        None.
        
        Examples
        --------
        
        >>> test_list=[24,26,75,124,220]
        # first define a dict object that define some protein features.

       
        >>> example_1=Annotator(protein_length=250, figure_size=(5,3), figure_dpi=200)
        # creating a VisTool instance

       
        >>> example_1.add_base_track()
        # add a base_track
        
        >>> example_1.add_marked_positions_track(test_list) # build a marked position track using the default parameters
        # marked positions track

        
        >>> example_1.add_marked_positions_track(positions=test_list,height_frac=0.75,
                                          track_label="Post_translational_modifications",
                                          marker_bar_dict={"color":"blue"})
        # add a second marked position track with the following parameters:
        #track name:  Post_translational_modifications
        #hight of the maker bar = 75%
        #color of the markerbar= blue

        Notes
        ------
        Any panel can have zero, one or more than one marked-position track.
        Thus, in the above examples calling the method ``add_marked_positions_track`` 
        for the second time does NOT override the previous marked-position track 
        it create a new one and added to the figure.
        """
        # adjust the current Geometry of the track.
        number_axes=len(self.fig.axes)
        for i in range(number_axes):
            self.fig.axes[i].change_geometry(number_axes+1,1,i+1)
        # create a new axes
        ax=self.fig.add_subplot(number_axes+1,1,number_axes+1)
        # adjust the axis size
        height=1
        ax.axis(xmin=0,xmax=self.protein_length,ymin=0,ymax=height)
        # draw a base line
        ax.hlines(y=0,xmin=0,xmax=self.protein_length,**base_line_dict)
        # plotting the figure        
        for pos in positions:
            ax.vlines(x=pos,ymin=0,ymax=height_frac,
                      **marker_bar_dict)
            ax.scatter(x=pos,y=height_frac,**marker_dict)
        # adjust the lay out of the figure
        ax.spines["left"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.spines["bottom"].set_visible(False)      
        # adjust the edges
        ax.get_xaxis().set_ticks([])
        ax.get_yaxis().set_ticks([])
        # add the label:
        ax.set_ylabel(track_label,track_label_dict)
    
    def add_coverage_track(self, coverage_matrix: np.ndarray,
                          coverage_as_base: bool =False,
                          coverage_dict: Dict[str,Union[int,str]]={"color":"blue","width":1.2},
                          xlabel:str="positions",
                          xlabel_dict:Dict[str,Union[int,str]]={"fontsize":6,"color":"black"},
                          ylabel:str ="coverage",
                          ylabel_dict:Dict[str,Union[int,str]]={"fontsize":6,"color":"black"},
                          number_ticks:int =10,
                          xticks_font_size: int =4,
                          yticks_font_size:int =4):
        """
        Description
        -----------
        Adds a coverage plot to the panel. The coverage plot shows the 
        relationship between a peptide and its experimentally detected eluted 
        peptide/peptides.

        Parameters
        ----------
        coverage_matrix : np.ndarray
            A protein length by one array which summarize information about the protein and 
            the eluted peptides.
            
        coverage_as_base : bool, optional
            Whether or not to plot the coverage as a base track for the figure.
            The default is False which means that the track appended to a figure 
            that have a default base track which can be constructed using the 
            method ``add_base_track``. However, if coverage_as_base is set to 
            True, the function will draw the base track using the coverage matrix and
            calling the function add_base_track should be avoided. 
            
        coverage_dict : Dict[str,Union[int,str]], optional
            The parameters that control the printing of the coverage matrix, 
            for example, the color. These parameters are fed to the function
            ``axes.bar``. The default is {"color":"blue","width":1.2}.
            
        xlabel : str, optional
            The label of the x-axis of the coverage track. The default is "positions".
            
        xlabel_dict : Dict[str,Union[int,str]], optional
            The parameters that control the x-label printing, for example,
            the color and/ the font size. these parameters are fed to the function
            ``axes.set_xlabel``. The default is {"fontsize":6,"color":"black"}.
            
        ylabel : str, optional
            The label of the y-axis of the coverage track. The default is "coverage".
            
        ylabel_dict : Dict[str,Union[int,str]], optional
            The parameters that control the x-label printing, for example,
            the color and/ the font size. these parameters are fed to the function
            ``axes.set_ylabel``. The default is {"fontsize":6,"color":"black"}.
            
        number_ticks : int, optional
            The number of ticks on the x-axis. The default is 10.
            
        xticks_font_size : float, optional
            The font size of the x-axis ticks. The default is 4.
            
        yticks_font_size : float, optional
            The font size of the y-axis ticks. The default is 4.

        """
        # adjust the figure axes:
        if coverage_as_base:
            number_axes=len(self.fig.axes)
            for i in range(number_axes):
                self.fig.axes[i].change_geometry(number_axes+1,1,i+1)
            # create a new axes
            ax=self.fig.add_subplot(number_axes+1,1,number_axes+1)
        else:
            ax=self.fig.add_subplot(111) # coverage is the base track
            
        # Adjust the limits
        heightest_coverage=max(coverage_matrix.reshape(-1))
        ax.axis(xmin=0,xmax=self.protein_length,ymin=0,
                ymax=heightest_coverage+0.1*heightest_coverage)
        
        # Plotting the matrix
        ax.bar(np.arange(coverage_matrix.shape[0]),
                          coverage_matrix.reshape(-1),**coverage_dict)
        # remove the spines of the figure
        ax.spines["left"].set_visible(False)   
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.spines["bottom"].set_visible(False)
        # adjust the x and y labels
        if not coverage_as_base:
            ax.get_xaxis().set_ticks([])
        else:
            ax.xaxis.tick_top()
        ax.set_ylabel(ylabel,**ylabel_dict)
        ax.set_xlabel(xlabel,xlabel_dict)
        step_size=int(np.floor(self.protein_length/number_ticks))
        ax.set_xticks(ticks=np.arange(0,self.protein_length,step_size),minor=False)
        # adjust the tick size
        ax.tick_params(axis='x', which='major', labelsize=xticks_font_size)
        ax.tick_params(axis='y', which='major', labelsize=yticks_font_size)
    
    def get_figure(self)->plt.Figure:
        """
        Returns
        -------
        matplotlib.figure.Figure
            The figure with all the tracks that have been added to it.
        """
        return self.fig
    
    def save_fig(self,name: str,
                output_path:str=".", 
                format_:str="png",
                figure_dpi:str="same",
                figure_saving_dict:Dict[str,Union[int,str]]={"facecolor":"white"})->None:
        """
        Description
        -----------
        Write the constructed figure to the disk.

        Parameters
        ----------
        name : str
            The name of the figure to save the file.
        
        output_path : str , optional
            The path to write the output, by default the function write to the
            current working directory.
        
        format_ : str, optional
            The output format, this parameter will be fed to the method
            ``plt.savefig``. The default is "png".
        
        figure_dpi: int, optional 
            The dpi of the saved figure. The deafult is same which means the figure 
            will be saved using the same dpi used for creating the figure. 
        
        figure_saving_dict: Dict[str,Union[int,str]],optional
            The parameters that should be fed to the function ``plt.savefig``.
            The default is figure_saving_dict={"facecolor":"white"}
        
        Returns
        -------
        None.

        """
        # Check the user input
        if figure_dpi=="same":
            figure_dpi=self.figure_dpi
        else:
            assert isinstance(figure_dpi,int), """
            figure_dpi must be an int object, or the string same for the figure current dpi"""
        assert isinstance(figure_saving_dict,dict), """ figure_saving_dict must be a dict object"""
        # figure name
        filename=os.path.join(output_path,name)
        # save the figure
        self.fig.savefig(fname=filename,dpi=figure_dpi,format=format_,**figure_saving_dict)




#!/usr/bin/env Python 
""" An abstraction for a collection of HLA alleles
"""
# load the modules 
from __future__ import annotations
import numpy as np 
from IPTK.Classes.HLAMolecules import HLAMolecule
from IPTK.Utils.Types import HLA_Names
from mhcnames import allele_name
from typing import List 
# define the class 
class HLASet: 
    def __init__(self, hlas: HLA_Names, gene_sep: str = ':')-> HLASet:
        """create an HLASet class which is a collection of HLA molecules 
        :param hlas: a list of alleles to be added to the set  
        :type hlas: HLA_Names
        :param gene_sep: Incase of HLA-DP and HLA-DQ, what is the gene separator that separate the genes' names, 
        for example, in DQA1*0303:DQB1*0202 the separator is the colon, defaults to ':'.
        :type gene_sep: str, optional
        :raises ValueError: incase the list of alleles is empty 
        :raises ValueError: incase of HLA-DP and HLA-DQ with a mismatch allele names.
        :raises RuntimeError: incase adding the alleles failed for any reason.
        :return: an HLASet instance 
        :rtype: HLASet
        """
        # check that the provided list is not empty 
        if len(hlas)==0:
            raise ValueError(f"The provided list is empty!") 
        # add the alleles to the dict of objects 
        self._hlas=dict()
        for hla in hlas: 
            try:
                if 'DQ' in hla or 'DP' in hla:
                    chains=hla.split(gene_sep)
                    if len(chains)>2:
                        raise ValueError(f'splitting the allele name: {hla} with separator: {gene_sep} generated {len(chains)} chains!!, at max two chain are allowed')
                    elif len(chains)==2:
                        self._hlas[hla]=HLAMolecule(chain_one=chains[0], chain_two=chains[1])
                    else: 
                        self._hlas[hla]=HLAMolecule(chain=hla[0])
                else: 
                    self._hlas[hla]=HLAMolecule(chain=hla)
            except Exception as exp: 
                raise RuntimeError(f'While adding the allele: {hla} to the set, the following error : {exp} was encountered.')
        # check that all alleles belong to the same class 
        self._assert_same_class()
        return 

    def get_hla_count(self)->int:
        """
        :return: The count of HLA molecules in the set 
        :rtype: int
        """
        return len(self)
    
    def _assert_same_class(self)->None:
        """
        :raises ValueError: if the alleles in the set are of different classes, for example,  HLA-I or HLA-II
        """
        # get the HLA-class 
        # get a radom class from the input HLA 
        default_class=self._hlas[list(self._hlas.keys())[0]].get_class() # get the class of a random molecules 
        # check that all the alleles belong to the same class 
        for allele_name in self._hlas.keys(): 
            if self._hlas[allele_name].get_class()!=default_class: 
                raise ValueError(f"ASSERTION FAILED: the alleles in the set belong to a different classes")
        
    def get_class(self)->int:
        """
        :return: The class of the HLA-alleles in the current instance 
        :rtype: int
        """
        return self._hlas[list(self._hlas.keys())[0]].get_class()
    
    def get_alleles(self)-> List[str]:
        """
        :return: The current alleles in the set
        :rtype: int
        """
        return list(self._hlas.keys())
    
    def has_allele(self,allele: str)->bool:
        """
        :param allele: The name of the alleles to check for its occurrence in the instance. 
        :type allele: str
        :return: True, if the provided allele is in the current instance, False otherwise. 
        :rtype: bool
        """
        allele_names=list(self._hlas.keys())
        for name in allele_names:
            if allele == name or allele == name.strip('HLA').strip('-') or allele.strip('HLA').strip('-') == name :
                return True
        return False
    
    def has_gene(self,gene_name:str)->bool:
        """
        :param gene_name: the gene name to search the set against.
        :type gene_name: str
        :return: True, if at least one of the alleles in the set belongs to the provided gene. False otherwise
        :rtype: bool
        """
        for allele_name in self._hlas.keys(): 
            if gene_name in self._hlas[allele_name].get_gene(): 
                return True 
        return False 
    
    def has_allele_group(self, allele_group:str)->bool:
        """
        :param allele_group: The allele group to search the set for 
        :type allele_group: str
        :return: True, if at least one allele in the set belongs to the provided allele group, False otherwise. 
        :rtype: bool
        """
        for allele_name in self._hlas.keys(): 
            if allele_group in self._hlas[allele_name].get_allele_group():
                return True 
        return False 
    
    def has_protein_group(self,protein_group:str)->bool:
        """
        :param protein_group: The protein group to search the set for
        :type protein_group: 
        :return: True, if at least one allele in the set belongs to the provided protein group
        :rtype: bool
        """
        for allele_name in self._hlas.keys(): 
            if protein_group in self._hlas[allele_name].get_protein_group():
                return True 
        return False 

    # define some magic functions
    def __len__(self)->int:
        """
        :return: The number of alleles in the set 
        :rtype: int
        """
        return len(self._hlas)

    def __str__(self) ->str:
        """
        :return: A string form of the class  
        :rtype: str
        """
        return f"An HLASet containing {len(self)} alleles"
    
    def __repr__(self)->str:
        return str(self)
            
            
            
            
            